import xtrack as xt

# Load a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

# Build optimizer object for tunes and chromaticities without performing optimization
opt = line.match(
    solve=False, # <--
    method='4d',
    vary=[
        xt.VaryList(['kqtf.b1', 'kqtd.b1'], step=1e-8, tag='quad'),
        xt.VaryList(['ksf.b1', 'ksd.b1'], step=1e-4, limits=[-0.1, 0.1], tag='sext'),
    ],
    targets = [
        xt.TargetSet(qx=62.315, qy=60.325, tol=1e-6, tag='tune'),
        xt.TargetSet(dqx=10.0, dqy=12.0, tol=0.01, tag='chrom'),
    ])

opt.target_status()
# prints the following (optimization not performed):
#
# Target status:
# id state tag   tol_met     residue current_val target_val description
#  0 ON    tune    False -0.00499997       62.31     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune    False      -0.005       60.32     60.325 'qy', val=60.325, tol=1e-06, weight=10)
#  2 ON    chrom   False    -8.09005     1.90995         10 'dqx', val=10, tol=0.01, weight=1)
#  3 ON    chrom   False     -10.057     1.94297         12 'dqy', val=12, tol=0.01, weight=1)

# Disable optimization of chromaticities and usage of sextupole knobs
opt.disable_targets(tag='chrom')
opt.disable_vary(tag='sext')

opt.show()
# prints:
#
# Vary:
# id tag  state description
#  0 quad ON    name='kqtf.b1', limits=None, step=1e-08, weight=1)
#  1 quad ON    name='kqtd.b1', limits=None, step=1e-08, weight=1)
#  2 sext OFF   name='ksf.b1', limits=(-0.1, 0.1), step=0.0001, weight=1)
#  3 sext OFF   name='ksd.b1', limits=(-0.1, 0.1), step=0.0001, weight=1)
# Targets:
# id tag   state description
#  0 tune  ON    'qx', val=62.315, tol=1e-06, weight=10)
#  1 tune  ON    'qy', val=60.325, tol=1e-06, weight=10)
#  2 chrom OFF   'dqx', val=10, tol=0.01, weight=1)
#  3 chrom OFF   'dqy', val=12, tol=0.01, weight=1)

# Solve (for tunes only)
opt.solve()

opt.target_status()
# prints:
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True -4.51905e-12      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune     True  9.23706e-14      60.325     60.325 'qy', val=60.325, tol=1e-06, weight=10)
#  2 OFF   chrom   False            0     1.89374         10 'dqx', val=10, tol=0.01, weight=1)
#  3 OFF   chrom   False            0     1.91882         12 'dqy', val=12, tol=0.01, weight=1)

# Enable all targets and knobs
opt.enable_all_targets()
opt.enable_all_vary()

# Solve (for tunes and chromaticities)
opt.solve()

opt.target_status()
# prints:
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True -5.62885e-10      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune     True  2.67875e-11      60.325     60.325 'qy', val=60.325, tol=1e-06, weight=10)
#  2 ON    chrom    True -0.000156234     9.99984         10 'dqx', val=10, tol=0.01, weight=1)
#  3 ON    chrom    True -9.81714e-07          12         12 'dqy', val=12, tol=0.01, weight=1)

# Change a target value and the corresponding tolerance
opt.targets[1].value = 60.05
opt.targets[1].tol = 1e-10

opt.target_status()
# prints:
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True -5.62885e-10      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune    False        0.275      60.325      60.05 'qy', val=60.05, tol=1e-10, weight=10)
#  2 ON    chrom    True -0.000156234     9.99984         10 'dqx', val=10, tol=0.01, weight=1)
#  3 ON    chrom    True -9.81714e-07          12         12 'dqy', val=12, tol=0.01, weight=1)

# Perform two optimization steps (without checking for convergence)
opt.step(2)

opt.target_status()
# prints (two steps were not enough to reach convergence):
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True  4.55631e-08      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune    False  2.56767e-09       60.05      60.05 'qy', val=60.05, tol=1e-10, weight=10)
#  2 ON    chrom    True -0.000127644     9.99987         10 'dqx', val=10, tol=0.01, weight=1)
#  3 ON    chrom    True -2.44325e-05          12         12 'dqy', val=12, tol=0.01, weight=1)

# Perform additional two steps
opt.step(2)

opt.target_status()
# prints (convergence was reached):
#
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True -4.00533e-11      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune     True  1.42109e-14       60.05      60.05 'qy', val=60.05, tol=1e-10, weight=10)
#  2 ON    chrom    True  3.19579e-07          10         10 'dqx', val=10, tol=0.01, weight=1)
#  3 ON    chrom    True -1.30694e-09          12         12 'dqy', val=12, tol=0.01, weight=1)

# Tag present configuration
opt.tag(tag='my_tag')

# Reload initial configuration
opt.reload(iteration=0)
opt.target_status()
# prints:
#
# Target status:
# id state tag   tol_met     residue current_val target_val description
#  0 ON    tune    False -0.00499997       62.31     62.315 'qx', val=62.315, tol=1e-06, weight=10
#  1 ON    tune    False        0.27       60.32      60.05 'qy', val=60.05, tol=1e-10, weight=10
#  2 ON    chrom   False    -8.09005     1.90995         10 'dqx', val=10, tol=0.01, weight=1
#  3 ON    chrom   False     -10.057     1.94297         12 'dqy', val=12, tol=0.01, weight=1

# Reload tagged configuration
opt.reload(tag='my_tag')
opt.target_status()
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True -4.00533e-11      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10
#  1 ON    tune     True  1.42109e-14       60.05      60.05 'qy', val=60.05, tol=1e-10, weight=10
#  2 ON    chrom    True  3.19579e-07          10         10 'dqx', val=10, tol=0.01, weight=1
#  3 ON    chrom    True -1.30694e-09          12         12 'dqy', val=12, tol=0.01, weight=1


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #
import math

import numpy as np
import pytest

import xobjects as xo
import xtrack as xt
from xtrack.slicing import Strategy, Teapot, Uniform, Custom, Slicer


def test_slicing_uniform():
    # Test for one slice
    slicing_1 = Uniform(1)
    assert slicing_1.element_weights() == [1.0]
    assert slicing_1.drift_weights() == [0.5] * 2
    expected_1 = [w for w in slicing_1.iter_weights(None)]
    result_1 = [(0.5, True), (1.0, False), (0.5, True)]
    assert expected_1 == result_1

    # Test for three slices
    slicing_3 = Uniform(3)
    assert slicing_3.element_weights() == [1/3] * 3
    assert slicing_3.drift_weights() == [0.25] * 4

    elem_info, drift_info = (1./3., False), (0.25, True)
    expected_3 = [
        drift_info, elem_info,
        drift_info, elem_info,
        drift_info, elem_info,
        drift_info,
    ]
    result_3 = [w for w in slicing_3.iter_weights(None)]
    assert expected_3 == result_3

    # Test error handling
    with pytest.raises(ValueError):
        Uniform(0)


def test_slicing_teapot():
    # Test for one slice
    slicing_1 = Teapot(1)
    assert slicing_1.element_weights() == [1.0]
    assert slicing_1.drift_weights() == [0.5] * 2
    expected_1 = [(0.5, True), (1.0, False), (0.5, True)]
    result_1 = [w for w in slicing_1.iter_weights()]
    assert expected_1 == result_1

    # Test for three slices
    slicing_3 = Teapot(3)
    assert slicing_3.element_weights() == [1/3] * 3
    assert slicing_3.drift_weights() == [0.125, 0.375, 0.375, 0.125]

    elem_info = (1./3., False)
    expected_3 = [
        (0.125, True), elem_info,
        (0.375, True), elem_info,
        (0.375, True), elem_info,
        (0.125, True),
    ]
    result_3 = [w for w in slicing_3.iter_weights()]
    assert expected_3 == result_3

    # Test error handling
    with pytest.raises(ValueError):
        Teapot(0)


def test_slicing_teapot_mode_thick():
    # Test for two slices
    slicing_2 = Teapot(2, mode='thick')
    assert slicing_2.drift_weights() == [0.5] * 2
    expected_2 = [(0.5, True), (0.5, True)]
    result_2 = [w for w in slicing_2.iter_weights()]
    assert expected_2 == result_2

    # Test for four slices
    slicing_3 = Teapot(4, mode='thick')
    assert slicing_3.drift_weights() == [0.125, 0.375, 0.375, 0.125]

    expected = [
        (0.125, True),
        (0.375, True),
        (0.375, True),
        (0.125, True),
    ]
    result = [w for w in slicing_3.iter_weights()]
    assert expected == result

    # Test error handling
    with pytest.raises(ValueError):
        Teapot(0)


def test_slicing_custom():
    elem_len_3 = 16
    slicing_3 = Custom(at_s=[0.8, 2, 8], mode='thin')
    expected_dr_3 = [0.8/16, 1.2/16, 6/16, 8/16]
    result_dr_3 = slicing_3.drift_weights(element_length=elem_len_3)
    xo.assert_allclose(expected_dr_3, result_dr_3, atol=1e-30)

    expected_el_3 = [1/3] * 3
    result_el_3 = slicing_3.element_weights(element_length=elem_len_3)
    xo.assert_allclose(expected_el_3, result_el_3, atol=1e-30)

    elem_info = (1/3, False)
    expected_3 = [
        (0.8/16, True), elem_info,
        (1.2/16, True), elem_info,
        (6/16, True), elem_info,
        (8/16, True),
    ]
    result_3 = [w for w in slicing_3.iter_weights(element_length=elem_len_3)]
    assert expected_3 == result_3  # ditto


def test_slicing_custom_thick():
    elem_len_1 = 1.1
    slicing_1 = Custom(at_s=[0.3], mode='thick')
    expected_dr_1 = [0.3 / 1.1, 0.8 / 1.1]
    result_dr_1 = slicing_1.drift_weights(element_length=elem_len_1)
    xo.assert_allclose(expected_dr_1, result_dr_1, atol=1e-30)

    expected_1 = [
        (0.3 / 1.1, True),
        (0.8 / 1.1, True),
    ]
    result_1 = [w for w in slicing_1.iter_weights(element_length=elem_len_1)]
    assert expected_1 == result_1  # for now exact comparison works

    elem_len_3 = 16
    slicing_3 = Custom(at_s=[0.8, 2, 8])
    expected_dr_3 = [0.8/16, 1.2/16, 6/16, 8/16]
    result_dr_3 = slicing_3.drift_weights(element_length=elem_len_3)
    xo.assert_allclose(expected_dr_3, result_dr_3, atol=1e-30)

    expected_3 = [
        (0.8/16, True),
        (1.2/16, True),
        (6/16, True),
        (8/16, True),
    ]
    result_3 = [w for w in slicing_3.iter_weights(element_length=elem_len_3)]
    assert expected_3 == result_3  # ditto


def test_strategy_matching_good_order():
    slicing_strategies = [
        Strategy(slicing=Uniform(1)),
        Strategy(element_type=xt.Drift, slicing=Uniform(2)),
        Strategy(name='some.*', slicing=Uniform(3)),
        Strategy(name='something', slicing=Uniform(4)),
        Strategy(name='something', element_type=xt.Drift, slicing=Uniform(5)),
    ]

    dr = xt.Drift()
    mk = xt.Marker()
    line = xt.Line(elements=[dr, mk])
    slicer = Slicer(slicing_strategies=slicing_strategies, line=line)

    assert slicer._scheme_for_element(element=mk, name='else', line=line).slicing_order == 1
    assert slicer._scheme_for_element(element=dr, name='what', line=line).slicing_order == 2
    assert slicer._scheme_for_element(element=dr, name='somewhat', line=line).slicing_order == 3
    assert slicer._scheme_for_element(element=mk, name='something', line=line).slicing_order == 4
    assert slicer._scheme_for_element(element=dr, name='something', line=line).slicing_order == 5


def test_strategy_matching_confusing_order():
    slicing_strategies = [
        Strategy(slicing=Uniform(1)),
        Strategy(name='something', slicing=Uniform(2)),
        Strategy(name='something', element_type=xt.Drift, slicing=Uniform(3)),
        Strategy(name='some.*', slicing=Uniform(4)),
        Strategy(name='some.*', element_type=xt.Drift, slicing=Uniform(5)),
    ]

    dr = xt.Drift()
    mk = xt.Marker()
    line = xt.Line(elements=[dr, mk])
    slicer = Slicer(slicing_strategies=slicing_strategies, line=line)

    assert slicer._scheme_for_element(element=mk, name='else', line=line).slicing_order == 1
    assert slicer._scheme_for_element(element=mk, name='something', line=line).slicing_order == 4
    assert slicer._scheme_for_element(element=dr, name='something', line=line).slicing_order == 5
    assert slicer._scheme_for_element(element=mk, name='somewhat', line=line).slicing_order == 4
    assert slicer._scheme_for_element(element=dr, name='somewhat', line=line).slicing_order == 5


def test_slicing_strategy_matching():
    elements = [
        ('keep_this', xt.Quadrupole(length=1.0)),
        ('mb10', xt.Bend(length=1.0)),
        ('keep_drifts', xt.Drift(length=1.0)),
        ('mb11', xt.Quadrupole(length=1.0)),
        ('mq10', xt.Quadrupole(length=1.0)),
        ('something', xt.Bend(length=1.0)),
        ('mb20', xt.Bend(length=1.0)),
        ('keep_thin', xt.Multipole(length=1.0)),
        ('mb21', xt.Quadrupole(length=1.0)),
    ]

    slicing_strategies = [
        # Default: one slice
        Strategy(slicing=Uniform(1)),
        # All bends: two slices
        Strategy(slicing=Teapot(2), element_type=xt.Bend),
        # All CFDs: three slices
        Strategy(slicing=Uniform(3), element_type=xt.Quadrupole),
        # If the name starts with mb: five slices (the bend and the cfd 'mb11')
        Strategy(slicing=Teapot(5), name=r'mb1.*'),
        # If the name starts with some: four slices (the bend 'something')
        Strategy(slicing=Uniform(4), name=r'some.*'),
        # Keep the CFD 'keep'
        Strategy(slicing=None, name=r'keep'),
        # mq10: one slice
        Strategy(slicing=Teapot(1), name=r'mq10'),
    ]

    line = xt.Line(
        elements=dict(elements),
        element_names=[name for name, _ in elements],
    )

    line.slice_thick_elements(slicing_strategies)

    # Check that the slices are as expected:
    expected_names = [
        # Kept Quadrupole:
        'keep_this',
        # 5 slices for mb10:
        'mb10_entry',  # Marker
        'mb10..entry_map',
        'drift_mb10..0', 'mb10..0', 'drift_mb10..1', 'mb10..1', 'drift_mb10..2',
        'mb10..2', 'drift_mb10..3', 'mb10..3', 'drift_mb10..4', 'mb10..4',
        'drift_mb10..5',
        'mb10..exit_map',
        'mb10_exit',  # Marker
        # Kept drift:
        'keep_drifts',
        # Five slices for mb11:
        'mb11_entry',  # Marker
        'mb11..entry_map',
        'drift_mb11..0', 'mb11..0', 'drift_mb11..1', 'mb11..1', 'drift_mb11..2',
        'mb11..2', 'drift_mb11..3', 'mb11..3', 'drift_mb11..4', 'mb11..4',
        'drift_mb11..5',
        'mb11..exit_map',
        'mb11_exit',  # Marker
        # One slice for 'mq10':
        'mq10_entry',  # Marker
        'mq10..entry_map',
        'drift_mq10..0', 'mq10..0', 'drift_mq10..1',
        'mq10..exit_map',
        'mq10_exit',  # Marker
        # Four slices for 'something':
        'something_entry',  # Marker
        'something..entry_map',
        'drift_something..0', 'something..0', 'drift_something..1',
        'something..1', 'drift_something..2', 'something..2',
        'drift_something..3', 'something..3', 'drift_something..4',
        'something..exit_map',
        'something_exit',  # Marker
        # Two slices for 'mb20':
        'mb20_entry',  # Marker
        'mb20..entry_map',
        'drift_mb20..0', 'mb20..0', 'drift_mb20..1', 'mb20..1', 'drift_mb20..2',
        'mb20..exit_map',
        'mb20_exit',  # Marker
        # Keep thin:
        'keep_thin',
        # Three slices for 'mb21' (it's a Quadrupole!):
        'mb21_entry',  # Marker
        'mb21..entry_map',
        'drift_mb21..0', 'mb21..0', 'drift_mb21..1', 'mb21..1', 'drift_mb21..2',
        'mb21..2', 'drift_mb21..3',
        'mb21..exit_map',
        'mb21_exit',  # Marker
    ]
    assert line.element_names == expected_names

    # Check types:
    for name, element in line.element_dict.items():
        if name == 'keep_this':
            assert isinstance(element, xt.Quadrupole)
        elif name == 'keep_drifts':
            assert isinstance(element, xt.Drift)
        elif name.startswith('drift_'):
            assert 'DriftSlice' in type(element).__name__
        elif name == 'keep_thin':
            assert isinstance(element, xt.Multipole)
        elif name[-3:-1] == '..':
            assert 'ThinSlice' in type(element).__name__
        else:
            name not in line.element_names

    # Check the right scheme was used:
    def _weights_of_drifts(name):
        return [
            line[nn].weight for nn in line.element_names
            if nn.startswith(f'drift_{name}')
        ]

    # Check markers
    assert isinstance(line['mb10_entry'], xt.Marker)
    assert isinstance(line['mb11_entry'], xt.Marker)
    assert isinstance(line['mq10_entry'], xt.Marker)
    assert isinstance(line['something_entry'], xt.Marker)
    assert isinstance(line['mb20_entry'], xt.Marker)
    assert isinstance(line['mb21_entry'], xt.Marker)
    assert isinstance(line['mb10_exit'], xt.Marker)
    assert isinstance(line['mb11_exit'], xt.Marker)
    assert isinstance(line['mq10_exit'], xt.Marker)
    assert isinstance(line['something_exit'], xt.Marker)
    assert isinstance(line['mb20_exit'], xt.Marker)
    assert isinstance(line['mb21_exit'], xt.Marker)

    # Teapot
    expected_mb20_drift_lens = [1/6, 2/3, 1/6]
    assert _weights_of_drifts('mb20') == expected_mb20_drift_lens

    expected_mb10_drift_lens = [1/12, 5/24, 5/24, 5/24, 5/24, 1/12]
    assert _weights_of_drifts('mb10') == expected_mb10_drift_lens

    expected_mb11_drift_lens = [1/12, 5/24, 5/24, 5/24, 5/24, 1/12]
    assert _weights_of_drifts('mb11') == expected_mb11_drift_lens

    # Uniform
    expected_mq10_drift_lens = [1/2] * 2
    assert _weights_of_drifts('mq10') == expected_mq10_drift_lens

    expected_mb21_drift_lens = [1/4] * 4
    assert _weights_of_drifts('mb21') == expected_mb21_drift_lens

    expected_something_drift_lens = [1/5] * 5
    assert _weights_of_drifts('something') == expected_something_drift_lens

    tt = line.get_table()

    # Test accessing compound elements
    assert np.all(tt.rows['mb10_entry':'mb10_exit'].name == [
        'mb10_entry', 'mb10..entry_map',
        'drift_mb10..0', 'mb10..0', 'drift_mb10..1', 'mb10..1',
        'drift_mb10..2', 'mb10..2', 'drift_mb10..3', 'mb10..3', 'drift_mb10..4',
        'mb10..4', 'drift_mb10..5',
        'mb10..exit_map', 'mb10_exit',
    ])
    assert np.all(tt.rows['mb11_entry':'mb11_exit'].name == [
        'mb11_entry', 'mb11..entry_map',
        'drift_mb11..0', 'mb11..0', 'drift_mb11..1', 'mb11..1',
        'drift_mb11..2', 'mb11..2', 'drift_mb11..3', 'mb11..3', 'drift_mb11..4',
        'mb11..4', 'drift_mb11..5',
        'mb11..exit_map', 'mb11_exit',
    ])
    assert np.all(tt.rows['mq10_entry':'mq10_exit'].name == [
        'mq10_entry', 'mq10..entry_map',
        'drift_mq10..0', 'mq10..0', 'drift_mq10..1',
        'mq10..exit_map', 'mq10_exit',
    ])
    assert np.all(tt.rows['something_entry':'something_exit'].name == [
        'something_entry', 'something..entry_map',
        'drift_something..0', 'something..0',
        'drift_something..1', 'something..1', 'drift_something..2',
        'something..2', 'drift_something..3', 'something..3',
        'drift_something..4',
        'something..exit_map', 'something_exit',
    ])
    assert np.all(tt.rows['mb20_entry':'mb20_exit'].name == [
        'mb20_entry', 'mb20..entry_map',
        'drift_mb20..0', 'mb20..0', 'drift_mb20..1', 'mb20..1',
        'drift_mb20..2',
        'mb20..exit_map', 'mb20_exit',
    ])
    assert np.all(tt.rows['mb21_entry':'mb21_exit'].name == [
        'mb21_entry', 'mb21..entry_map',
        'drift_mb21..0', 'mb21..0', 'drift_mb21..1', 'mb21..1',
        'drift_mb21..2', 'mb21..2', 'drift_mb21..3',
        'mb21..exit_map', 'mb21_exit',
    ])


def test_slicing_thick_bend_simple():

    additional_kwargs = {}
    additional_kwargs['k1'] = 0.2

    bend = xt.Bend(
        length=3.0,
        k0=0.1,
        h=0.2,
        **additional_kwargs
    )
    line = xt.Line(elements=[bend], element_names=['bend'])
    line.slice_thick_elements([Strategy(slicing=Teapot(2))])

    assert len(line) == 9  # 2 markers + 2 edges + 2 slices + 3 drifts

    assert line['drift_bend..0'].weight == 1/6
    assert line['drift_bend..1'].weight == 2/3
    assert line['drift_bend..2'].weight == 1/6

    bend0, bend1 = line['bend..0'], line['bend..1']
    assert bend0.weight == bend1.weight == 0.5

    # Make sure the order and the inverse factorial make sense:
    _fact = math.factorial
    xo.assert_allclose(_fact(bend0._parent.order) * bend0._parent.inv_factorial_order, 1, atol=1e-16)
    xo.assert_allclose(_fact(bend1._parent.order) * bend0._parent.inv_factorial_order, 1, atol=1e-16)


def test_slicing_thick_bend_into_thick_bends_simple():

    additional_kwargs = {}
    additional_kwargs['k0'] = 0.1
    additional_kwargs['h'] = 0.2
    additional_kwargs['k1'] = 0.2

    bend = xt.Bend(
        length=3.0,
        knl=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        ksl=[0.7, 0.6, 0.5, 0.4, 0.3, 0.2],
        **additional_kwargs
    )
    line = xt.Line(elements=[bend], element_names=['bend'])
    line.slice_thick_elements([Strategy(slicing=Uniform(2, mode='thick'))])

    assert len(line) == 6  # 2 markers + 2 edges + 2 thick slices

    bend0, bend1 = line['bend..0'], line['bend..1']
    assert bend0.weight == bend1.weight == 0.5

    assert bend0._parent.k0 == bend1._parent.k0 == 0.1
    assert bend0._parent.h == bend1._parent.h == 0.2
    assert bend0._parent.k1 == bend1._parent.k1 == 0.2

    expected_knl = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
    xo.assert_allclose(bend0._parent.knl, expected_knl, atol=1e-16)
    xo.assert_allclose(bend1._parent.knl, expected_knl, atol=1e-16)

    expected_ksl = np.array([0.7, 0.6, 0.5, 0.4, 0.3, 0.2])
    xo.assert_allclose(bend0._parent.ksl, expected_ksl, atol=1e-16)
    xo.assert_allclose(bend1._parent.ksl, expected_ksl, atol=1e-16)

    # Make sure the order and the inverse factorial make sense:
    _fact = math.factorial
    xo.assert_allclose(_fact(bend0._parent.order) * bend0._parent.inv_factorial_order, 1, atol=1e-16)
    xo.assert_allclose(_fact(bend1._parent.order) * bend0._parent.inv_factorial_order, 1, atol=1e-16)


def test_slicing_xdeps_consistency():
    num_elements = 50000
    num_slices = 1

    line = xt.Line(
        elements=[xt.Bend(k0=1, length=100)] * num_elements,
        element_names=[f'bend{ii}' for ii in range(num_elements)],
    )
    line._init_var_management()

    for ii in range(num_elements):
        line.vars[f'k{ii}'] = 1
        line.element_refs[f'bend{ii}'].k0 = line.vars[f'k{ii}']

    sgy = xt.slicing.Strategy(
        element_type=xt.Bend,
        slicing=xt.slicing.Uniform(num_slices),
    )
    line.slice_thick_elements([sgy])
    assert len(line.to_dict()['_var_manager']) == num_elements * num_slices

def test_slice_twice():
    env = xt.Environment()
    line = env.new_line(components=[
        env.new('el', xt.Bend, length=4, k1=0.5, h=0.1, k0=0.8, knl=[0, 0, 0.03]),
    ])

    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(slicing=xt.Teapot(4, mode='thick'), element_type=xt.Bend),
        ])

    tt_first_slice = line.get_table(attr=True)

    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(slicing=xt.Teapot(4, mode='thin'), element_type=xt.ThickSliceBend),
        ])

    tt_second_slice = line.get_table(attr=True)

    assert np.all(tt_first_slice.name == np.array(
        ['el_entry', 'el..entry_map', 'el..0', 'el..1', 'el..2', 'el..3',
        'el..exit_map', 'el_exit', '_end_point']))

    assert np.all(tt_first_slice.element_type == np.array(
        ['Marker', 'ThinSliceBendEntry', 'ThickSliceBend', 'ThickSliceBend',
        'ThickSliceBend', 'ThickSliceBend', 'ThinSliceBendExit', 'Marker',
        '']))

    # Check first table

    xo.assert_allclose(tt_first_slice.angle_rad, np.array(
        [0.  , 0.  , 0.05, 0.15, 0.15, 0.05, 0.  , 0.  , 0.  ]), rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_first_slice.angle_rad), 0.4, rtol=0, atol=1e-12)

    xo.assert_allclose(tt_first_slice.k0l, np.array(
        [0. , 0. , 0.4, 1.2, 1.2, 0.4, 0. , 0. , 0. ]), rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_first_slice.k0l), 3.2, rtol=0, atol=1e-12)

    xo.assert_allclose(tt_first_slice.k1l, np.array(
        [0.  , 0.  , 0.25, 0.75, 0.75, 0.25, 0.  , 0.  , 0.  ]), rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_first_slice.k1l), 2, rtol=0, atol=1e-12)

    xo.assert_allclose(tt_first_slice.k2l, np.array(
        [0.     , 0.     , 0.00375, 0.01125, 0.01125, 0.00375, 0. , 0., 0.]),
        rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_first_slice.k2l), 0.03, rtol=0, atol=1e-12)

    # Check second table

    assert np.all(tt_second_slice.name == np.array(
        ['el_entry', 'el..entry_map', 'el..0_entry', 'drift_el..0..0',
        'el..0..0', 'drift_el..0..1', 'el..0..1', 'drift_el..0..2',
        'el..0..2', 'drift_el..0..3', 'el..0..3', 'drift_el..0..4',
        'el..0_exit', 'el..1_entry', 'drift_el..1..0', 'el..1..0',
        'drift_el..1..1', 'el..1..1', 'drift_el..1..2', 'el..1..2',
        'drift_el..1..3', 'el..1..3', 'drift_el..1..4', 'el..1_exit',
        'el..2_entry', 'drift_el..2..0', 'el..2..0', 'drift_el..2..1',
        'el..2..1', 'drift_el..2..2', 'el..2..2', 'drift_el..2..3',
        'el..2..3', 'drift_el..2..4', 'el..2_exit', 'el..3_entry',
        'drift_el..3..0', 'el..3..0', 'drift_el..3..1', 'el..3..1',
        'drift_el..3..2', 'el..3..2', 'drift_el..3..3', 'el..3..3',
        'drift_el..3..4', 'el..3_exit', 'el..exit_map', 'el_exit',
        '_end_point']))

    assert np.all(tt_second_slice.element_type == np.array(
        ['Marker', 'ThinSliceBendEntry', 'Marker', 'DriftSliceBend',
        'ThinSliceBend', 'DriftSliceBend', 'ThinSliceBend',
        'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend',
        'ThinSliceBend', 'DriftSliceBend', 'Marker', 'Marker',
        'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend',
        'ThinSliceBend', 'DriftSliceBend', 'ThinSliceBend',
        'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend', 'Marker',
        'Marker', 'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend',
        'ThinSliceBend', 'DriftSliceBend', 'ThinSliceBend',
        'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend', 'Marker',
        'Marker', 'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend',
        'ThinSliceBend', 'DriftSliceBend', 'ThinSliceBend',
        'DriftSliceBend', 'ThinSliceBend', 'DriftSliceBend', 'Marker',
        'ThinSliceBendExit', 'Marker', ''
        ]))

    xo.assert_allclose(tt_second_slice.angle_rad, np.array(
        [0.    , 0.    , 0.    , 0.    , 0.0125, 0.    , 0.0125, 0.    ,
            0.0125, 0.    , 0.0125, 0.    , 0.    , 0.    , 0.    , 0.0375,
            0.    , 0.0375, 0.    , 0.0375, 0.    , 0.0375, 0.    , 0.    ,
            0.    , 0.    , 0.0375, 0.    , 0.0375, 0.    , 0.0375, 0.    ,
            0.0375, 0.    , 0.    , 0.    , 0.    , 0.0125, 0.    , 0.0125,
            0.    , 0.0125, 0.    , 0.0125, 0.    , 0.    , 0.    , 0.    ,
            0.    ]), rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_second_slice.angle_rad), 0.4, rtol=0, atol=1e-12)

    xo.assert_allclose(tt_second_slice.k0l, np.array(
        [0. , 0. , 0. , 0. , 0.1, 0. , 0.1, 0. , 0.1, 0. , 0.1, 0. , 0. ,
        0. , 0. , 0.3, 0. , 0.3, 0. , 0.3, 0. , 0.3, 0. , 0. , 0. , 0. ,
        0.3, 0. , 0.3, 0. , 0.3, 0. , 0.3, 0. , 0. , 0. , 0. , 0.1, 0. ,
        0.1, 0. , 0.1, 0. , 0.1, 0. , 0. , 0. , 0. , 0. ]), rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_second_slice.k0l), 3.2, rtol=0, atol=1e-12)

    xo.assert_allclose(tt_second_slice.k1l, np.array(
        [0.    , 0.    , 0.    , 0.    , 0.0625, 0.    , 0.0625, 0.    ,
        0.0625, 0.    , 0.0625, 0.    , 0.    , 0.    , 0.    , 0.1875,
        0.    , 0.1875, 0.    , 0.1875, 0.    , 0.1875, 0.    , 0.    ,
        0.    , 0.    , 0.1875, 0.    , 0.1875, 0.    , 0.1875, 0.    ,
        0.1875, 0.    , 0.    , 0.    , 0.    , 0.0625, 0.    , 0.0625,
        0.    , 0.0625, 0.    , 0.0625, 0.    , 0.    , 0.    , 0.    ,
        0.    ]), rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_second_slice.k1l), 2, rtol=0, atol=1e-12)

    xo.assert_allclose(tt_second_slice.k2l, np.array(
        [0.       , 0.       , 0.       , 0.       , 0.0009375, 0.       ,
        0.0009375, 0.       , 0.0009375, 0.       , 0.0009375, 0.       ,
        0.       , 0.       , 0.       , 0.0028125, 0.       , 0.0028125,
        0.       , 0.0028125, 0.       , 0.0028125, 0.       , 0.       ,
        0.       , 0.       , 0.0028125, 0.       , 0.0028125, 0.       ,
        0.0028125, 0.       , 0.0028125, 0.       , 0.       , 0.       ,
        0.       , 0.0009375, 0.       , 0.0009375, 0.       , 0.0009375,
        0.       , 0.0009375, 0.       , 0.       , 0.       , 0.       ,
        0.       ]),
        rtol=0, atol=1e-12)
    xo.assert_allclose(np.sum(tt_second_slice.k2l), 0.03, rtol=0, atol=1e-12)

def test_slice_repeated_elements():

    env = xt.Environment(
        particle_ref=xt.Particles(p0c=7000e9, x=1e-3, px=1e-3, y=1e-3, py=1e-3))

    line0 = env.new_line(
        components=[
            env.new('b0', 'Bend', length=1.0, anchor='start', at=5.0),
            env.new('q0', 'Quadrupole', length=2.0, at=20.0),
            env.new('ql', 'Quadrupole', length=2.0, at=-10.0, from_='q0'),
            env.new('qr', 'Quadrupole', length=2.0, at=10.0, from_='q0'),
            env.new('mk1', 'Marker', at=40),
            env.new('mk2', 'Marker'),
            env.new('mk3', 'Marker'),
            env.place('q0'),
            env.place('b0'),
            env.new('end', 'Marker', at=50.),
        ])
    tt0 = line0.get_table()
    tt0.show(cols=['name', 's_start', 's_end', 's_center'])

    line = line0.copy()
    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(slicing=xt.Teapot(2), name=r'q0.*'),
            xt.Strategy(slicing=xt.Teapot(3), name=r'b0.*'),
        ]
    )
    tt = line.get_table()
    tt.show(cols=['name', 's_start', 's_end', 's_center'])

    assert np.all(tt.name == np.array(
        ['drift_1', 'b0_entry::0', 'b0..entry_map', 'drift_b0..0', 'b0..0',
       'drift_b0..1', 'b0..1', 'drift_b0..2', 'b0..2', 'drift_b0..3',
       'b0..exit_map', 'b0_exit::0', 'drift_2', 'ql', 'drift_3',
       'q0_entry::0', 'q0..entry_map', 'drift_q0..0', 'q0..0',
       'drift_q0..1', 'q0..1', 'drift_q0..2', 'q0..exit_map',
       'q0_exit::0', 'drift_4', 'qr', 'drift_5', 'mk1', 'mk2', 'mk3',
       'q0_entry::1', 'q0..entry_map_0', 'drift_q0..3', 'q0..2',
       'drift_q0..4', 'q0..3', 'drift_q0..5', 'q0..exit_map_0',
       'q0_exit::1', 'b0_entry::1', 'b0..entry_map_0', 'drift_b0..4',
       'b0..3', 'drift_b0..5', 'b0..4', 'drift_b0..6', 'b0..5',
       'drift_b0..7', 'b0..exit_map_0', 'b0_exit::1', 'drift_6', 'end',
       '_end_point']))
    xo.assert_allclose(tt.s_center, np.array(
        [ 2.5       ,  5.        ,  5.        ,  5.0625    ,  5.125     ,
        5.3125    ,  5.5       ,  5.6875    ,  5.875     ,  5.9375    ,
        6.        ,  6.        ,  7.5       , 10.        , 15.        ,
       19.        , 19.        , 19.16666667, 19.33333333, 20.        ,
       20.66666667, 20.83333333, 21.        , 21.        , 25.        ,
       30.        , 35.5       , 40.        , 40.        , 40.        ,
       40.        , 40.        , 40.16666667, 40.33333333, 41.        ,
       41.66666667, 41.83333333, 42.        , 42.        , 42.        ,
       42.        , 42.0625    , 42.125     , 42.3125    , 42.5       ,
       42.6875    , 42.875     , 42.9375    , 43.        , 43.        ,
       46.5       , 50.        , 50.        ]),
        rtol=0., atol=1e-8)

    line = line0.copy()
    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(slicing=xt.Teapot(2), name=r'q0::0'),
            xt.Strategy(slicing=xt.Teapot(3), name=r'b0::1'),
        ]
    )
    tt = line.get_table()
    tt.show(cols=['name', 's_start', 's_end', 's_center'])

    assert np.all(tt.name == np.array(
        ['drift_1', 'b0', 'drift_2', 'ql', 'drift_3', 'q0_entry',
       'q0..entry_map', 'drift_q0..0', 'q0..0', 'drift_q0..1', 'q0..1',
       'drift_q0..2', 'q0..exit_map', 'q0_exit', 'drift_4', 'qr',
       'drift_5', 'mk1', 'mk2', 'mk3', 'q0', 'b0_entry', 'b0..entry_map',
       'drift_b0..0', 'b0..0', 'drift_b0..1', 'b0..1', 'drift_b0..2',
       'b0..2', 'drift_b0..3', 'b0..exit_map', 'b0_exit', 'drift_6',
       'end', '_end_point']))

    xo.assert_allclose(tt.s_center, np.array(
        [ 2.5       ,  5.5       ,  7.5       , 10.        , 15.        ,
          19.        , 19.        , 19.16666667, 19.33333333, 20.        ,
          20.66666667, 20.83333333, 21.        , 21.        , 25.        ,
          30.        , 35.5       , 40.        , 40.        , 40.        ,
          41.        , 42.        , 42.        , 42.0625    , 42.125     ,
          42.3125    , 42.5       , 42.6875    , 42.875     , 42.9375    ,
          43.        , 43.        , 46.5       , 50.        , 50.        ]),
        rtol=0., atol=1e-8)

    line = line0.copy()
    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(slicing=xt.Teapot(2, mode='thick'), name=r'q0.*'),
            xt.Strategy(slicing=xt.Teapot(3, mode='thick'), name=r'b0.*'),
        ],
    )

    tt = line.get_table()
    tt.show(cols=['name', 's_start', 's_end', 's_center'])

    assert np.all(tt.name == np.array(
        ['drift_1', 'b0_entry::0', 'b0..entry_map', 'b0..0', 'b0..1',
       'b0..2', 'b0..exit_map', 'b0_exit::0', 'drift_2', 'ql', 'drift_3',
       'q0_entry::0', 'q0..entry_map', 'q0..0', 'q0..1', 'q0..exit_map',
       'q0_exit::0', 'drift_4', 'qr', 'drift_5', 'mk1', 'mk2', 'mk3',
       'q0_entry::1', 'q0..entry_map_0', 'q0..2', 'q0..3',
       'q0..exit_map_0', 'q0_exit::1', 'b0_entry::1', 'b0..entry_map_0',
       'b0..3', 'b0..4', 'b0..5', 'b0..exit_map_0', 'b0_exit::1',
       'drift_6', 'end', '_end_point']))
    xo.assert_allclose(tt.s_center, np.array(
        [ 2.5       ,  5.        ,  5.        ,  5.08333333,  5.5       ,
          5.91666667,  6.        ,  6.        ,  7.5       , 10.        ,
          15.        , 19.        , 19.        , 19.5       , 20.5       ,
          21.        , 21.        , 25.        , 30.        , 35.5       ,
          40.        , 40.        , 40.        , 40.        , 40.        ,
          40.5       , 41.5       , 42.        , 42.        , 42.        ,
          42.        , 42.08333333, 42.5       , 42.91666667, 43.        ,
          43.        , 46.5       , 50.        , 50.        ]),
        rtol=0., atol=1e-8)

    line = line0.copy()
    line.cut_at_s([20.1, 20.2, 41.7, 41.8, 5.5])

    tt = line.get_table()
    tt.show(cols=['name', 's_start', 's_end', 's_center'])

    assert np.all(tt.name == np.array(
        ['drift_1', 'b0_entry', 'b0..entry_map', 'b0..0', 'b0..1',
       'b0..exit_map', 'b0_exit', 'drift_2', 'ql', 'drift_3',
       'q0_entry::0', 'q0..entry_map', 'q0..0', 'q0..1', 'q0..2',
       'q0..exit_map', 'q0_exit::0', 'drift_4', 'qr', 'drift_5', 'mk1',
       'mk2', 'mk3', 'q0_entry::1', 'q0..entry_map_0', 'q0..3', 'q0..4',
       'q0..5', 'q0..exit_map_0', 'q0_exit::1', 'b0', 'drift_6', 'end',
       '_end_point']))

    xo.assert_allclose(tt.s_center, np.array(
        [ 2.5 ,  5.  ,  5.  ,  5.25,  5.75,  6.  ,  6.  ,  7.5 , 10.  ,
       15.  , 19.  , 19.  , 19.55, 20.15, 20.6 , 21.  , 21.  , 25.  ,
       30.  , 35.5 , 40.  , 40.  , 40.  , 40.  , 40.  , 40.85, 41.75,
       41.9 , 42.  , 42.  , 42.5 , 46.5 , 50.  , 50.  ]),
        rtol=0., atol=1e-8)



import numpy as np
import xtrack as xt


tilt_machine_by_90_degrees = True
wiggler_on = True
vertical_orbit_distortion = True
check_against_tracking = True

line = xt.Line.from_json('fccee_h_thin.json')
line.build_tracker()

print('Done building tracker')

if wiggler_on:
    line.vars['on_wiggler_v'] = 0.4

if vertical_orbit_distortion:
    line['mwi.e5rg'].knl[0] = 2e-7

if tilt_machine_by_90_degrees:

    tw_before_tilt = line.twiss()

    # Bring the machine to the vertical plane
    already_tilted = []
    for nn in line.element_names:
        if hasattr(line[nn], 'rot_s_rad'):
            if nn not in already_tilted:
                line[nn].rot_s_rad += np.pi/2
                already_tilted.append(nn)
        if hasattr(line[nn], 'parent_name'):
            nn_parent = line[nn].parent_name
            if hasattr(line[nn_parent], 'rot_s_rad'):
                if nn_parent not in already_tilted:
                    line[nn_parent].rot_s_rad += np.pi/2
                    already_tilted.append(nn_parent)

    tw_after_tilt = line.twiss()

    assert_allclose = np.testing.assert_allclose
    assert_allclose(tw_after_tilt.qy, tw_before_tilt.qx, rtol=0, atol=1e-8)
    assert_allclose(tw_after_tilt.qx, tw_before_tilt.qy, rtol=0, atol=1e-8)
    assert_allclose(tw_after_tilt.dqy, tw_before_tilt.dqx, rtol=0, atol=1e-4)
    assert_allclose(tw_after_tilt.dqx, tw_before_tilt.dqy, rtol=0, atol=1e-4)

    assert_allclose(tw_after_tilt.bety, tw_before_tilt.betx, rtol=1e-5, atol=0)
    assert_allclose(tw_after_tilt.betx, tw_before_tilt.bety, rtol=1e-5, atol=0)

    assert_allclose(tw_after_tilt.y, tw_before_tilt.x, rtol=0, atol=1e-9)
    assert_allclose(tw_after_tilt.x, -tw_before_tilt.y, rtol=0, atol=1e-9)

    assert_allclose(tw_after_tilt.dy, tw_before_tilt.dx, rtol=0, atol=5e-6)
    assert_allclose(tw_after_tilt.dx, -tw_before_tilt.dy, rtol=0, atol=5e-6)

tw_no_rad = line.twiss(method='4d')

line.configure_radiation(model='mean')
line.compensate_radiation_energy_loss()

tw_rad = line.twiss(eneloss_and_damping=True)
ex = tw_rad.eq_gemitt_x
ey = tw_rad.eq_gemitt_y
ez = tw_rad.eq_gemitt_zeta



import numpy as np
from scipy.constants import c as clight
from scipy.constants import hbar
from scipy.constants import epsilon_0

from cpymad.madx import Madx
import xtrack as xt
import xobjects as xo

mad = Madx()
mad.call('../../test_data/sps_thick/sps.seq')

# mad.input('beam, particle=proton, pc=26;')
# mad.input('beam, particle=electron, pc=20;')

# # realistic
# mad.input('beam, particle=electron, pc=20;')
# v_mv = 25
# num_turns = 8000

# higher energy
mad.input('beam, particle=electron, pc=50;')
v_mv = 250
num_turns = 600

mad.call('../../test_data/sps_thick/lhc_q20.str')

mad.use(sequence='sps')

mad.input('twiss, table=tw4d;')
twm4d = mad.table.tw4d

n_cav = 6

mad.sequence.sps.elements['actcse.31632'].volt = v_mv * 10 / n_cav   # To stay in the linear region
mad.sequence.sps.elements['actcse.31632'].freq = 3
mad.sequence.sps.elements['actcse.31632'].lag = 0.5


mad.input('twiss, table=tw6d;')
twm6d = mad.table.tw6d

mad.sequence.sps.beam.radiate = True
mad.emit()

line = xt.Line.from_madx_sequence(mad.sequence.sps, allow_thick=True,
                                  deferred_expressions=True)
line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                                    q0=-1, gamma0=mad.sequence.sps.beam.gamma)
line.cycle('bpv.11706', inplace=True)

line.insert_element(element=line['actcse.31632'].copy(), index='bpv.11706',
                    name='cav1')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.21508',
                    name='cav2')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.41508',
                    name='cav4')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.51508',
                    name='cav5')
line.insert_element(element=line['actcse.31632'].copy(), index='bpv.61508',
                    name='cav6')

tt = line.get_table()

s_start_wig = tt['s', 'actcsg.31780']

line.discard_tracker()
line.insert_element(name='wig1', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+0*0.5)
line.insert_element(name='wig2', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+1*0.5)
line.insert_element(name='wig3', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+2*0.5)
line.insert_element(name='wig4', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+3*0.5)
line.insert_element(name='wig5', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+4*0.5)
line.insert_element(name='wig6', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+5*0.5)
line.insert_element(name='wig7', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+6*0.5)
line.insert_element(name='wig8', element=xt.Bend(length=0.5, rot_s_rad=np.pi/2), at_s=s_start_wig+0.1+7*0.5)

line['k_wig'] = 0
line['wig1'].k0 = 'k_wig'
line['wig2'].k0 = '-k_wig'
line['wig3'].k0 = '-k_wig'
line['wig4'].k0 = 'k_wig'
line['wig5'].k0 = '-k_wig'
line['wig6'].k0 = 'k_wig'
line['wig7'].k0 = 'k_wig'
line['wig8'].k0 = '-k_wig'

line['k_wig'] = 1e-3
tw = line.twiss4d()

# Remove edge effects
# for nn in tt.rows[tt.element_type=='DipoleEdge'].name:
#     line[nn].k = 0

tw_thick = line.twiss()

Strategy = xt.slicing.Strategy
Teapot = xt.slicing.Teapot

line.discard_tracker()
slicing_strategies = [
    Strategy(slicing=Teapot(1)),  # Default
    Strategy(slicing=Teapot(2), element_type=xt.Bend),
    Strategy(slicing=Teapot(8), element_type=xt.Quadrupole),
    Strategy(slicing=Teapot(20), name='wig.*'),
]

line.slice_thick_elements(slicing_strategies)
line.build_tracker()

line.discard_tracker()
line.build_tracker(_context=xo.ContextCpu())
line.configure_radiation(model=None)

line.configure_radiation(model='mean')
twr = line.twiss(eneloss_and_damping=True)



import pathlib

import numpy as np

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_ip_knob_matching_new_optimize_api(test_context):

    collider = xt.Environment.from_json(test_data_folder /
                    'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(test_context)

    # Add default steps and limits

    limitmcbxh =  63.5988e-6
    limitmcbxv =  67.0164e-6
    limitmcbx  =  67.0164e-6
    limitmcby  =  96.3000e-6
    limitmcb   =  80.8000e-6
    limitmcbc  =  89.8700e-6
    limitmcbw  =  80.1400e-6
    nrj = 7000.
    scale = 23348.89927*0.9
    scmin = 0.03*7000./nrj

    collider.vars.vary_default.update({
        # Correctors in IR8 (Q4 and Q5, and triplet)
        'acbyhs4.l8b1':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs4.r8b2':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs4.l8b2':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs4.r8b1':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbchs5.l8b2':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbchs5.l8b1':{'step': 1.0e-15,'limits':(-limitmcby, limitmcby)},
        'acbyhs5.r8b1':{'step': 1.0e-15,'limits':(-limitmcbc, limitmcbc)},
        'acbyhs5.r8b2':{'step': 1.0e-15,'limits':(-limitmcbc, limitmcbc)},
        'acbxh1.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh2.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh3.l8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh1.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh2.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
        'acbxh3.r8':   {'step': 1.0e-15, 'limits':(-limitmcbx, limitmcbx)},
    })

    # Check a few steps and limits

    xo.assert_allclose(collider.vars.vary_default['acbxh1.l8']['step'], 1.0e-15, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbxh1.l8']['limits'][0], -limitmcbx, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbxh1.l8']['limits'][1], limitmcbx, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbyhs5.r8b2']['step'], 1.0e-15, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbyhs5.r8b2']['limits'][0], -limitmcbc, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbyhs5.r8b2']['limits'][1], limitmcbc, atol=1e-17, rtol=0)

    # Check that they are preserved by to_dict/from_dict
    collider = xt.Environment.from_dict(collider.to_dict())
    collider.build_trackers()

    xo.assert_allclose(collider.vars.vary_default['acbxh1.l8']['step'], 1.0e-15, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbxh1.l8']['limits'][0], -limitmcbx, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbxh1.l8']['limits'][1], limitmcbx, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbyhs5.r8b2']['step'], 1.0e-15, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbyhs5.r8b2']['limits'][0], -limitmcbc, atol=1e-17, rtol=0)
    xo.assert_allclose(collider.vars.vary_default['acbyhs5.r8b2']['limits'][1], limitmcbc, atol=1e-17, rtol=0)

    # kill all existing knobs orbit knobs in ip2 and ip8
    all_knobs_ip2ip8 = [
        'acbxh3.r2', 'acbchs5.r2b1', 'pxip2b1', 'acbxh2.l8',
        'acbyhs4.r8b2', 'pyip2b1', 'acbxv1.l8', 'acbyvs4.l2b1', 'acbxh1.l8',
        'acbxv2.r8', 'pxip8b2', 'yip8b1', 'pxip2b2', 'acbcvs5.r2b1', 'acbyhs4.l8b1',
        'acbyvs4.l8b1', 'acbxh2.l2', 'acbxh3.l2', 'acbxv1.r8', 'acbxv1.r2',
        'acbyvs4.r2b2', 'acbyvs4.l2b2', 'yip8b2', 'xip2b2', 'acbxh2.r2',
        'acbyhs4.l2b2', 'acbxv2.r2', 'acbyhs5.r8b1', 'acbxh2.r8', 'acbxv3.r8',
        'acbyvs5.r8b2', 'acbyvs5.l2b2', 'yip2b1', 'acbxv2.l2', 'acbyhs4.r2b2',
        'acbyhs4.r2b1', 'xip8b2', 'acbyvs5.l2b1', 'acbyvs4.r8b1', 'acbyvs4.r8b2',
        'acbyvs5.r8b1', 'acbxh1.r8', 'acbyvs4.l8b2', 'acbyhs5.l2b1', 'acbyvs4.r2b1',
        'acbcvs5.r2b2', 'acbcvs5.l8b2', 'acbyhs4.r8b1', 'pxip8b1', 'acbxv1.l2',
        'yip2b2', 'acbyhs4.l8b2', 'acbxv3.r2', 'xip8b1', 'acbchs5.r2b2', 'acbxh3.l8',
        'acbxh3.r8', 'acbyhs5.r8b2', 'acbxv2.l8', 'acbxh1.l2', 'pyip8b1', 'pyip8b2',
        'acbxv3.l8', 'xip2b1', 'acbyhs5.l2b2', 'acbchs5.l8b2', 'acbcvs5.l8b1',
        'pyip2b2', 'acbxv3.l2', 'acbchs5.l8b1', 'acbyhs4.l2b1', 'acbxh1.r2']

    for kk in all_knobs_ip2ip8:
        collider.vars[kk] = 0

    # Match horizontal xing angle in ip8
    angle_match = 300e-6
    opt = collider.match_knob(
        run=False,
        default_tol={'x': 1.1e-10, 'px': 0.9e-10},
        knob_name='on_x8h',
        knob_value_start=0,
        knob_value_end=(angle_match * 1e6),
        start=['s.ds.l8.b1', 's.ds.l8.b2'],
        end=['e.ds.r8.b1', 'e.ds.r8.b2'],
        init=[xt.TwissInit(), xt.TwissInit()],
        targets=[
            xt.TargetSet(x=0, px=0, at=xt.END, line='lhcb1'),
            xt.TargetSet(x=0, px=0, at=xt.END, line='lhcb2'),
            xt.Target('x', 0, at='ip8', line='lhcb1'),
            xt.Target('px', angle_match, at='ip8', line='lhcb1', tol=0.8e-10),
            xt.TargetSet(x=0, px=-angle_match, at='ip8', line='lhcb2'),
        ],
        vary=[
            # Vary with custom step and limits
            xt.VaryList(['acbyhs4.l8b1'], step=2e-15, limits=(-9e-5, 9e-5)),
            # Vary using default step and limits
            xt.VaryList([
                'acbyhs4.r8b2', 'acbyhs4.l8b2', 'acbyhs4.r8b1',
                'acbchs5.l8b2', 'acbchs5.l8b1', 'acbyhs5.r8b1', 'acbyhs5.r8b2']),
            xt.VaryList([
                'acbxh1.l8', 'acbxh2.l8', 'acbxh3.l8',
                'acbxh1.r8', 'acbxh2.r8', 'acbxh3.r8'], tag='mcbx')]
        )

    ll = opt.log()
    assert len(ll) == 1
    assert ll.iteration[0] == 0
    xo.assert_allclose(ll['penalty', 0], 0.0424264, atol=1e-6, rtol=0)
    assert ll['tol_met', 0] == 'yyyyynyn'
    assert ll['target_active', 0] == 'yyyyyyyy'
    assert ll['vary_active', 0] == 'yyyyyyyyyyyyyy'

    vnames = [vv.name for vv in opt.vary]
    assert np.all(np.array(vnames) == np.array(
    ['acbyhs4.l8b1_from_on_x8h', 'acbyhs4.r8b2_from_on_x8h',
    'acbyhs4.l8b2_from_on_x8h', 'acbyhs4.r8b1_from_on_x8h',
    'acbchs5.l8b2_from_on_x8h', 'acbchs5.l8b1_from_on_x8h',
    'acbyhs5.r8b1_from_on_x8h', 'acbyhs5.r8b2_from_on_x8h',
    'acbxh1.l8_from_on_x8h', 'acbxh2.l8_from_on_x8h', 'acbxh3.l8_from_on_x8h',
    'acbxh1.r8_from_on_x8h', 'acbxh2.r8_from_on_x8h', 'acbxh3.r8_from_on_x8h']))

    vtags = [vv.tag for vv in opt.vary]
    assert np.all(np.array(vtags) == np.array(
        ['', '', '', '', '', '', '', '', 'mcbx', 'mcbx', 'mcbx', 'mcbx', 'mcbx', 'mcbx']))

    xo.assert_allclose(opt.vary[0].step, 2e-15, atol=1e-17, rtol=0)
    xo.assert_allclose(opt.vary[0].limits[0], -9e-5, atol=1e-10, rtol=0)
    xo.assert_allclose(opt.vary[0].limits[1], 9e-5, atol=1e-10, rtol=0)
    xo.assert_allclose(opt.vary[2].step, 1e-15, atol=1e-17, rtol=0)
    xo.assert_allclose(opt.vary[2].limits[0], -limitmcby, atol=1e-10, rtol=0)
    xo.assert_allclose(opt.vary[2].limits[1], limitmcby, atol=1e-10, rtol=0)

    xo.assert_allclose(opt.targets[4].tol, 1.1e-10, atol=1e-14, rtol=0)
    xo.assert_allclose(opt.targets[5].tol, 0.8e-10, atol=1e-14, rtol=0)
    xo.assert_allclose(opt.targets[6].tol, 1.1e-10, atol=1e-14, rtol=0)
    xo.assert_allclose(opt.targets[7].tol, 0.9e-10, atol=1e-14, rtol=0)

    # Set mcmbx by hand (as in mad-x script)
    testkqx8=abs(collider.varval['kqx.l8'])*7000./0.3
    if testkqx8> 210.:
        acbx_xing_ir8 = 1.0e-6   # Value for 170 urad crossing
    else:
        acbx_xing_ir8 = 11.0e-6  # Value for 170 urad crossing

    # # As in mad-x script
    # collider.vars['acbxh1.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match / 170e-6
    # collider.vars['acbxh2.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match / 170e-6
    # collider.vars['acbxh3.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match / 170e-6
    # collider.vars['acbxh1.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match / 170e-6
    # collider.vars['acbxh2.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match / 170e-6
    # collider.vars['acbxh3.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match / 170e-6

    # Harder case to check mcbx matching
    collider.vars['acbxh1.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match / 170e-6 * 0.1
    collider.vars['acbxh2.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match / 170e-6 * 0.1
    collider.vars['acbxh3.l8_from_on_x8h'] = acbx_xing_ir8 * angle_match / 170e-6 * 0.1
    collider.vars['acbxh1.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match / 170e-6 * 0.1
    collider.vars['acbxh2.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match / 170e-6 * 0.1
    collider.vars['acbxh3.r8_from_on_x8h'] = -acbx_xing_ir8 * angle_match / 170e-6 * 0.1

    init_mcbx_plus = collider.varval['acbxh1.l8_from_on_x8h']

    # First round of optimization without changing mcbx
    opt.disable(vary='mcbx')
    vtags = [vv.tag for vv in opt.vary]
    assert np.all(np.array(vtags) == np.array(
        ['', '', '', '', '', '', '', '', 'mcbx', 'mcbx', 'mcbx', 'mcbx', 'mcbx', 'mcbx']))
    vactive = [vv.active for vv in opt.vary]
    assert np.all(np.array(vactive) == np.array(
        [True, True, True, True, True, True, True, True, False, False, False, False, False, False]))

    opt.step(10) # perform 10 steps without checking for convergence

    ll = opt.log()
    assert 11 <= len(ll) <= 13
    assert ll['vary_active', 0] == 'yyyyyyyyyyyyyy'
    assert ll['vary_active', 1] == 'yyyyyyyynnnnnn'
    assert ll['vary_active', len(ll) - 1] == 'yyyyyyyynnnnnn'

    # Check solution not found
    assert ll['tol_met', len(ll) - 1] != 'yyyyyyyy'

    # Check that mcbxs did not move
    xo.assert_allclose(ll['vary_8', 1:], init_mcbx_plus, atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_9', 1:], init_mcbx_plus, atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_10', 1:], init_mcbx_plus, atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_11', 1:], -init_mcbx_plus, atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_12', 1:], -init_mcbx_plus, atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_13', 1:], -init_mcbx_plus, atol=1e-12, rtol=0)


    # Link all mcbx stengths to the first one
    collider.vars['acbxh2.l8_from_on_x8h'] =  collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh3.l8_from_on_x8h'] =  collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh2.r8_from_on_x8h'] = -collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh3.r8_from_on_x8h'] = -collider.vars['acbxh1.l8_from_on_x8h']
    collider.vars['acbxh1.r8_from_on_x8h'] = -collider.vars['acbxh1.l8_from_on_x8h']

    # Enable first mcbx knob (which controls the others)
    assert opt.vary[8].name == 'acbxh1.l8_from_on_x8h'
    opt.vary[8].active = True

    opt.solve()
    ll = opt.log()

    # Check driving knob is enabled
    assert np.all(ll['vary_active', 13:] == 'yyyyyyyyynnnnn')

    # Check solution found
    assert np.all(ll['tol_met', -1] == 'yyyyyyyy')

    # Check imposed relationship among varys
    xo.assert_allclose(ll['vary_8', 13], ll['vary_9',   13], atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_8', 13], ll['vary_10',  13], atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_8', 13], -ll['vary_11', 13], atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_8', 13], -ll['vary_12', 13], atol=1e-12, rtol=0)
    xo.assert_allclose(ll['vary_8', 13], -ll['vary_13', 13], atol=1e-12, rtol=0)

    opt.generate_knob()

    collider.vars['on_x8h'] = 100
    tw = collider.twiss()
    collider.vars['on_x8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 100e-6, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -100e-6, atol=1e-10, rtol=0)

    # Match horizontal separation in ip8
    sep_match = 2e-3
    opt = collider.match_knob(
        run=False,
        knob_name='on_sep8h',
        knob_value_start=0,
        knob_value_end=(sep_match * 1e3),
        start=['s.ds.l8.b1', 's.ds.l8.b2'],
        end=['e.ds.r8.b1', 'e.ds.r8.b2'],
        init=[xt.TwissInit(betx=1, bety=1, element_name='s.ds.l8.b1', line=collider.lhcb1),
                    xt.TwissInit(betx=1, bety=1, element_name='s.ds.l8.b2', line=collider.lhcb2)],
        targets=[
            xt.TargetList(['x', 'px'], at='e.ds.r8.b1', line='lhcb1', value=0),
            xt.TargetList(['x', 'px'], at='e.ds.r8.b2', line='lhcb2', value=0),
            xt.Target('x', sep_match, at='ip8', line='lhcb1'),
            xt.Target('x', -sep_match, at='ip8', line='lhcb2'),
            xt.Target('px', 0, at='ip8', line='lhcb1'),
            xt.Target('px', 0, at='ip8', line='lhcb2'),
        ],
        vary=[
            xt.VaryList([
                'acbyhs4.l8b1', 'acbyhs4.r8b2', 'acbyhs4.l8b2', 'acbyhs4.r8b1',
                'acbchs5.l8b2', 'acbchs5.l8b1', 'acbyhs5.r8b1', 'acbyhs5.r8b2']),
            xt.VaryList([
                'acbxh1.l8', 'acbxh2.l8', 'acbxh3.l8',
                'acbxh1.r8', 'acbxh2.r8', 'acbxh3.r8'], tag='mcbx')]
        )

    # Set mcmbx by hand (as in mad-x script)
    testkqx8 = abs(collider.varval['kqx.l8'])*7000./0.3
    if testkqx8 > 210.:
        acbx_sep_ir8 = 18e-6   # Value for 170 urad crossing
    else:
        acbx_sep_ir8 = 16e-6  # Value for 170 urad crossing

    collider.vars['acbxh1.l8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3
    collider.vars['acbxh2.l8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3
    collider.vars['acbxh3.l8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3
    collider.vars['acbxh1.r8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3
    collider.vars['acbxh2.r8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3
    collider.vars['acbxh3.r8_from_on_sep8h'] = acbx_sep_ir8 * sep_match / 2e-3

    # First round of optimization without changing mcbx
    opt.disable(vary='mcbx')
    opt.step(10) # perform 10 steps without checking for convergence

    # Enable first mcbx knob (which controls the others)
    assert opt.vary[8].name == 'acbxh1.l8_from_on_sep8h'
    opt.vary[8].active = True

    opt.solve()
    opt.generate_knob()

    collider.vars['on_sep8h'] = 1.5
    tw = collider.twiss()
    collider.vars['on_sep8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 1.5e-3, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], -1.5e-3, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 0, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], 0, atol=1e-9, rtol=0)

    # Check that on_x8h still works
    collider.vars['on_x8h'] = 100
    tw = collider.twiss()
    collider.vars['on_x8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 100e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -100e-6, atol=1e-9, rtol=0)

    # Both knobs together
    collider.vars['on_x8h'] = 120
    collider.vars['on_sep8h'] = 1.7
    tw = collider.twiss()
    collider.vars['on_x8h'] = 0
    collider.vars['on_sep8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 1.7e-3, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], -1.7e-3, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 120e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -120e-6, atol=1e-9, rtol=0)

@for_all_test_contexts
def test_match_ir8_optics_new_optimize_api(test_context):

    collider = xt.Environment.from_json(test_data_folder /
                    'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(test_context)

    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1['betx', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip1'], 0.15, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip5'], 0.15, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip8'], 1.5, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip2'], 10., atol=1e-5, rtol=0)


    nrj = 7000.
    scale = 23348.89927*0.9
    scmin = 0.03*7000./nrj
    qtlimitx28 = 1.0*225.0/scale
    qtlimitx15 = 1.0*205.0/scale
    qtlimit2 = 1.0*160.0/scale
    qtlimit3 = 1.0*200.0/scale
    qtlimit4 = 1.0*125.0/scale
    qtlimit5 = 1.0*120.0/scale
    qtlimit6 = 1.0*90.0/scale

    collider.vars.vary_default.update({
        'kq4.r8b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq5.r8b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, -qtlimit2*scmin)},
        'kq6.r8b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq7.r8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq8.r8b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.r8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq10.r8b1':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.r8b1': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kq4.l8b1':    {'step': 1.0E-6, 'limits': (-qtlimit2, -qtlimit2*scmin)},
        'kq5.l8b1':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.l8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq7.l8b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.l8b1':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq9.l8b1':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.l8b1':   {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kqtl11.l8b1': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.l8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l8b1':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kq4.r8b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, -qtlimit2*scmin)},
        'kq5.r8b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.r8b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, -qtlimit2*scmin)},
        'kq7.r8b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq8.r8b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq9.r8b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq10.r8b2':   {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kqtl11.r8b2': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.r8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.r8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kq5.l8b2':    {'step': 1.0E-6, 'limits': (-qtlimit2, -qtlimit2*scmin)},
        'kq4.l8b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq6.l8b2':    {'step': 1.0E-6, 'limits': ( qtlimit2*scmin, qtlimit2)},
        'kq7.l8b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq8.l8b2':    {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kq9.l8b2':    {'step': 1.0E-6, 'limits': (-qtlimit3, -qtlimit3*scmin)},
        'kq10.l8b2':   {'step': 1.0E-6, 'limits': ( qtlimit3*scmin, qtlimit3)},
        'kqtl11.l8b2': {'step': 1.0E-6, 'limits': (-qtlimit4, qtlimit4)},
        'kqt12.l8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
        'kqt13.l8b2':  {'step': 1.0E-6, 'limits': (-qtlimit5, qtlimit5)},
    })


    mux_b1_target = 3.0985199176526272
    muy_b1_target = 2.7863674079923726

    mux_b2_target = 3.007814449420657
    muy_b2_target = 2.878419154545405

    collider.varval['kq6.l8b1'] *= 1.1
    collider.varval['kq6.r8b1'] *= 1.1

    tab_boundary_right = collider.lhcb1.twiss(
        start='ip8', end='ip1.l1',
        init=xt.TwissInit(element_name='ip1.l1', line=collider.lhcb1,
                                betx=0.15, bety=0.15))
    tab_boundary_left = collider.lhcb1.twiss(
        start='ip5', end='ip8',
        init=xt.TwissInit(element_name='ip5', line=collider.lhcb1,
                                betx=0.15, bety=0.15))

    opt = collider[f'lhcb1'].match(
        default_tol={None: 1e-7, 'betx': 1e-6, 'bety': 1e-6},
        solve=False,
        start=f's.ds.l8.b1', end=f'e.ds.r8.b1', init_at=xt.START,
        # Left boundary
        init=tab_boundary_left,
        targets=[
            xt.TargetSet(at='ip8', betx=1.5, bety=1.5, alfx=0, alfy=0, dx=0, dpx=0),
            xt.TargetSet(at=f'e.ds.r8.b1',
                         tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                         value=tab_boundary_right, tag='stage2'),
            xt.TargetRelPhaseAdvance('mux', mux_b1_target),
            xt.TargetRelPhaseAdvance('muy', muy_b1_target),
        ],
        vary=[
            xt.VaryList([
                f'kq6.l8b1', f'kq7.l8b1',
                f'kq8.l8b1', f'kq9.l8b1', f'kq10.l8b1', f'kqtl11.l8b1',
                f'kqt12.l8b1', f'kqt13.l8b1']
            ),
            xt.Vary(f'kq4.l8b1', tag='stage1', step=1.1e-6, limits=(-0.007, -0.00022)),
            xt.Vary(f'kq5.l8b1', tag='stage1'),
            xt.VaryList([
                f'kq4.r8b1', f'kq5.r8b1', f'kq6.r8b1', f'kq7.r8b1',
                f'kq8.r8b1', f'kq9.r8b1', f'kq10.r8b1', f'kqtl11.r8b1',
                f'kqt12.r8b1', f'kqt13.r8b1'],
                tag='stage2')
        ]
    )

    # Initial knob values
    init_knob_vals = opt.get_knob_values()

    assert opt.vary[8].name == 'kq4.l8b1'
    assert opt.vary[8].tag == 'stage1'
    assert opt.vary[8].step == 1.1e-6
    assert opt.vary[8].limits[0] == -0.007
    assert opt.vary[8].limits[1] == -0.00022

    assert opt.vary[9].name == 'kq5.l8b1'
    assert opt.vary[9].tag == 'stage1'
    assert opt.vary[9].step == 1.0e-6
    assert opt.vary[9].limits[0] == qtlimit2*scmin
    assert opt.vary[9].limits[1] == qtlimit2

    # Assert that the two quad break the optics
    assert opt.log()['tol_met', 0] == 'nnnnnnnnnnnnnn'

    # Check that the unperturbed machine is a solution
    collider.varval['kq6.l8b1'] /= 1.1
    collider.varval['kq6.r8b1'] /= 1.1

    opt.clear_log()
    assert opt.log()['tol_met', 0] == 'yyyyyyyyyyyyyy'

    # Break again and clear log
    collider.varval['kq6.l8b1'] *= 1.1
    collider.varval['kq6.r8b1'] *= 1.1
    opt.clear_log()
    assert opt.log()['tol_met', 0] == 'nnnnnnnnnnnnnn'

    opt.disable(target=['stage1', 'stage2'])
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyyyyyyyyyyyy'

    opt.disable(vary=['stage1', 'stage2'])
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyynnnnnnnnnnnn'

    opt.step(12)
    assert opt.log()['penalty', -1] < 0.1
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyynnnnnnnnnnnn'

    opt.enable(vary='stage1')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'

    opt.enable(target='stage1')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'

    opt.solve()
    assert opt.log()['penalty', -1] < 1e-7
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'

    opt.enable(target='stage2')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyyyyyyyyyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'
    assert opt.log()['tol_met', -1] != 'yyyyyyyyyyyyyy'

    opt.enable(vary='stage2')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyyyyyyyyyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyyyyyyyyyyyy'

    opt.solve()
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyyyyyyyyyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyyyyyyyyyyyy'
    assert opt.log()['tol_met', -1] == 'yyyyyyyyyyyyyy'


    xo.assert_allclose(tw.lhcb1['betx', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip1'], 0.15, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip5'], 0.15, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip8'], 1.5, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip2'], 10., atol=1e-5, rtol=0)

    # Beam 2

    collider.varval['kq6.l8b2'] *= 1.1
    collider.varval['kq6.r8b2'] *= 1.1

    tab_boundary_right = collider.lhcb2.twiss(
        start='ip8', end='ip1.l1',
        init=xt.TwissInit(element_name='ip1.l1', line=collider.lhcb2,
                                betx=0.15, bety=0.15))
    tab_boundary_left = collider.lhcb2.twiss(
        start='ip5', end='ip8',
        init=xt.TwissInit(element_name='ip5', line=collider.lhcb2,
                                betx=0.15, bety=0.15))

    opt = collider[f'lhcb2'].match(
        default_tol={None: 1e-7, 'betx': 5e-6, 'bety': 5e-6},
        solve=False,
        start=f's.ds.l8.b2', end=f'e.ds.r8.b2', init_at=xt.START,
        # Left boundary
        init=tab_boundary_left,
        targets=[
            xt.TargetSet(at='ip8', betx=1.5, bety=1.5, alfx=0, alfy=0, dx=0, dpx=0),
            xt.TargetSet(at=xt.END,
                         tars=('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                         value=tab_boundary_right, tag='stage2'),
            xt.TargetRelPhaseAdvance('mux', mux_b2_target),
            xt.TargetRelPhaseAdvance('muy', muy_b2_target),
        ],
        vary=[
            xt.VaryList([
                f'kq6.l8b2', f'kq7.l8b2',
                f'kq8.l8b2', f'kq9.l8b2', f'kq10.l8b2', f'kqtl11.l8b2',
                f'kqt12.l8b2', f'kqt13.l8b2']
            ),
            xt.Vary(f'kq4.l8b2', tag='stage1', step=1.1e-6, limits=(0.00022, 0.007)),
            xt.Vary(f'kq5.l8b2', tag='stage1'),
            xt.VaryList([
                f'kq4.r8b2', f'kq5.r8b2', f'kq6.r8b2', f'kq7.r8b2',
                f'kq8.r8b2', f'kq9.r8b2', f'kq10.r8b2', f'kqtl11.r8b2',
                f'kqt12.r8b2', f'kqt13.r8b2'],
                tag='stage2')
        ]
    )

    assert opt.vary[8].name == 'kq4.l8b2'
    assert opt.vary[8].tag == 'stage1'
    assert opt.vary[8].step == 1.1e-6
    assert opt.vary[8].limits[0] == 0.00022
    assert opt.vary[8].limits[1] == 0.007

    assert opt.vary[9].name == 'kq5.l8b2'
    assert opt.vary[9].tag == 'stage1'
    assert opt.vary[9].step == 1.0e-6
    assert opt.vary[9].limits[0] == -qtlimit2
    assert opt.vary[9].limits[1] == -qtlimit2*scmin

    # Assert that the two quad break the optics
    assert opt.log()['tol_met', 0] == 'nnnnnnnnnnnnnn'

    # Check that the unperturbed machine is a solution
    collider.varval['kq6.l8b2'] /= 1.1
    collider.varval['kq6.r8b2'] /= 1.1

    opt.clear_log()
    assert opt.log()['tol_met', 0] == 'yyyyyyyyyyyyyy'

    # Break again and clear log
    collider.varval['kq6.l8b2'] *= 1.1
    collider.varval['kq6.r8b2'] *= 1.1
    opt.clear_log()
    assert opt.log()['tol_met', 0] == 'nnnnnnnnnnnnnn'

    opt.disable(target=['stage1', 'stage2'])
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyyyyyyyyyyyy'

    opt.disable(vary=['stage1', 'stage2'])
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyynnnnnnnnnnnn'

    # Tag present state
    knob_values_before_tag0 = opt.get_knob_values()
    i_iter_tag0 = opt.log().iteration[-1]

    opt.step(10)
    assert opt.log()['penalty', -1] < 0.1
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyynnnnnnnnnnnn'

    # Tag present state
    knob_values_before_tag1 = opt.get_knob_values()
    opt.tag(tag='mytag1')

    ##### Check of reloading features #####

    # Check that knobs have changed
    assert np.any([knob_values_before_tag0[k] != knob_values_before_tag1[k]
                     for k in knob_values_before_tag0])

    # Reload with iteration number
    opt.reload(iteration=i_iter_tag0)
    knobs_now = opt.get_knob_values()
    assert np.all([knobs_now[k] == knob_values_before_tag0[k]
                        for k in knob_values_before_tag0])

    # Reload with tag
    opt.reload(tag='mytag1')
    knobs_now = opt.get_knob_values()
    assert np.all([knobs_now[k] == knob_values_before_tag1[k]
                        for k in knob_values_before_tag1])

    ##### Done checking reloading features #####

    opt.enable(vary='stage1')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'

    opt.enable(target='stage1')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'

    opt.solve()
    assert opt.log()['penalty', -1] < 1e-5
    assert opt.log()['target_active', -1] == 'yyyyyynnnnnnyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'

    opt.enable(target='stage2')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyyyyyyyyyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyynnnnnnnnnn'
    assert opt.log()['tol_met', -1] != 'yyyyyyyyyyyyyy'

    opt.enable(vary='stage2')
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyyyyyyyyyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyyyyyyyyyyyy'

    opt.solve()
    opt.tag()
    assert opt.log()['target_active', -1] == 'yyyyyyyyyyyyyy'
    assert opt.log()['vary_active', -1] == 'yyyyyyyyyyyyyyyyyyyy'
    assert opt.log()['tol_met', -1] == 'yyyyyyyyyyyyyy'

    xo.assert_allclose(tw.lhcb1['betx', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip1'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip1'], 0.15, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip5'], 0.15, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip5'], 0.15, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip8'], 1.5, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip8'], 1.5, atol=1e-6, rtol=0)

    xo.assert_allclose(tw.lhcb1['betx', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb1['bety', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb2['betx', 'ip2'], 10., atol=1e-5, rtol=0)
    xo.assert_allclose(tw.lhcb2['bety', 'ip2'], 10., atol=1e-5, rtol=0)




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import logging                            #!skip-doc
import numpy as np

import xtrack as xt
import xobjects as xo

# Display debug information .         #!skip-doc
logger = logging.getLogger('xtrack')  #!skip-doc
logger.setLevel(logging.DEBUG)        #!skip-doc

###################
# Build test line #
###################

ctx = xo.context_default
buf = ctx.new_buffer()

# We build a test line having two aperture elements which are shifted and
# rotated w.r.t. the accelerator reference frame.

# Define aper_0
aper_0 = xt.LimitEllipse(_buffer=buf, a=2e-2, b=1e-2)
shift_aper_0 = (1e-2, 0.5e-2)
rot_deg_aper_0 = 10.

# Define aper_1
aper_1 = xt.LimitRect(_buffer=buf, min_x=-1e-2, max_x=1e-2,
                                   min_y=-2e-2, max_y=2e-2)
shift_aper_1 = (-5e-3, 1e-2)
rot_deg_aper_1 = 10.
aper_1.shift_x = shift_aper_1[0]
aper_1.shift_y = shift_aper_1[1]
aper_1.rot_s_rad = np.deg2rad(rot_deg_aper_1)


# aper_0_sandwitch
line_aper_0 = xt.Line(
    elements=[xt.XYShift(_buffer=buf, dx=shift_aper_0[0], dy=shift_aper_0[1]),
              xt.SRotation(_buffer=buf, angle=rot_deg_aper_0),
              aper_0,
              xt.Multipole(_buffer=buf, knl=[0.001]),
              xt.SRotation(_buffer=buf, angle=-rot_deg_aper_0),
              xt.XYShift(_buffer=buf, dx=-shift_aper_0[0], dy=-shift_aper_0[1])])
line_aper_0.build_tracker(_buffer=buf)

# aper_1_sandwitch
line_aper_1 = xt.Line(
    elements=[aper_1,
              xt.Multipole(_buffer=buf, knl=[0.001])
        ])
line_aper_1.build_tracker(_buffer=buf)

#################
# Build tracker #
#################

line=xt.Line(
    elements = ((xt.Drift(_buffer=buf, length=0.5),)
                + line_aper_0.elements
                + (xt.Drift(_buffer=buf, length=1),
                   xt.Drift(_buffer=buf, length=1),
                   xt.Drift(_buffer=buf, length=1.),)
                + line_aper_1.elements))
line.build_tracker(_buffer=buf)
num_elements = len(line.element_names)

# Generate test particles
particles = xt.Particles(_context=ctx,
            px=np.random.uniform(-0.01, 0.01, 10000),
            py=np.random.uniform(-0.01, 0.01, 10000))

#########
# Track #
#########

line.track(particles)

########################
# Refine loss location #
########################

loss_loc_refinement = xt.LossLocationRefinement(line,
        n_theta = 360, # Angular resolution in the polygonal approximation of the aperture
        r_max = 0.5, # Maximum transverse aperture in m
        dr = 50e-6, # Transverse loss refinement accuracy [m]
        ds = 0.1, # Longitudinal loss refinement accuracy [m]
        save_refine_lines=True # Diagnostics flag
        )

import time                                              #!skip-doc
t0 = time.time()                                         #!skip-doc

loss_loc_refinement.refine_loss_location(particles)

t1 = time.time()                                         #!skip-doc
print(f'Took\t{(t1-t0)*1e3:.2f} ms')                     #!skip-doc


#!end-doc-part

import matplotlib.pyplot as plt
plt.close('all')

# Visualize apertures
interp_line = loss_loc_refinement.refine_lines[
                                loss_loc_refinement.i_apertures[1]]
s0 = interp_line.s0
s1 = interp_line.s1
polygon_0 = interp_line.elements[0]
polygon_1 = interp_line.elements[-1]
for ii, (ln, poly) in enumerate(
                         zip([line_aper_0, line_aper_1],
                             [polygon_0, polygon_1])):
    part_gen_range = 0.05
    pp = xt.Particles(
                    p0c=6500e9,
                    x=np.random.uniform(-part_gen_range, part_gen_range, 10000),
                    y=np.random.uniform(-part_gen_range, part_gen_range, 10000))
    x0 = pp.x.copy()
    y0 = pp.y.copy()

    ln.track(pp)
    ids = pp.particle_id


    fig = plt.figure(ii+1)
    ax = fig.add_subplot(111)
    plt.plot(x0, y0, '.', color='red')
    plt.axis('equal')
    plt.grid(linestyle=':')
    plt.plot(x0[ids][pp.state>0], y0[ids][pp.state>0], '.', color='green')
    plt.plot(poly.x_closed, poly.y_closed, '-k', linewidth=1)

plt.figure()
ax = plt.axes(projection='3d')
ax.plot3D(
        polygon_0.x_closed,
        polygon_0.y_closed,
        s0+polygon_0.x_closed*0,
        color='k', linewidth=3)
ax.plot3D(
        polygon_1.x_closed,
        polygon_1.y_closed,
        s1+polygon_1.x_closed*0,
        color='k', linewidth=3)
s_check = []
for ee, ss in zip(interp_line.elements,
                  interp_line.get_s_elements()):
    if ee.__class__ is xt.LimitPolygon:
        ax.plot3D(
                ee.x_closed,
                ee.y_closed,
                s0+ss+0*ee.x_closed,
                alpha=0.9,
                )
        s_check.append(ss)
mask = particles.at_element == loss_loc_refinement.i_apertures[1]
ax.plot3D(particles.x[mask], particles.y[mask], particles.s[mask],
          '.r', markersize=.5)
ax.view_init(65, 62); plt.draw()
ax.set_xlabel('x [m]')
ax.set_ylabel('y [m]')
ax.set_zlabel('s [m]')
plt.show()



import xtrack as xt

line = xt.Line.from_json('psb_investigate.json')

bb = line['bi1.bsw1l1.1']

rb = xt.RBend(length_straight=0.313, k0=0.211, k1=0, h=0*1e-6,
              k0_from_h=False, model='rot-kick-rot',
              num_multipole_kicks=1,
              knl=[ 0.      ,  0.      , -0.097429],
              edge_entry_active=0, edge_exit_active=0)

sb = xt.Bend(length=0.313, k0=0.211, k1=0, h=0*1e-6,
              k0_from_h=False,
              model='rot-kick-rot',
              num_multipole_kicks=1,
              knl=[ 0.      ,  0.      , -0.097429],
              edge_entry_active=0, edge_exit_active=0)

p0 = line.particle_ref.copy()

lrb = xt.Line(elements=[rb])
lrb.build_tracker()

prb = p0.copy()
psb = p0.copy()

lrb.track(prb)
sb.track(psb)

print(prb.x - psb.x)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import matplotlib.pyplot as plt

# https://gitlab.cern.ch/e-beam/pybetacool
from pybetacool import PyBetaCool

# unofficial Betacool: https://github.com/dgamba/betacool
obj = PyBetaCool(BLDfilename='LEIR.bld', betacoolExec='/home/pkruyt/cernbox/BETACOOL_ecooling/betacool/Betacool_Linux')

# Execute Betacool force
obj.runBetacool('/f')

#################################
# parse file and plot data
#################################

tmp = obj.parseCurveFile('flong.cur')
tmp = tmp[:(100000)]

v_diff=tmp[:, 0]
force=tmp[:,1]

plt.figure()
plt.plot(v_diff,-force, label='betacool')
plt.xlabel('delta v')
plt.ylabel('-Force [eV/m]')
plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
plt.legend()
plt.xlim([0,6e5])
plt.show()


#################################
# save data as reference for test
#################################
np.savez('../force_betacool.npz', v_diff=v_diff, force=force)


import numpy as np
from cpymad.madx import Madx
import xtrack as xt

from scipy.constants import c as clight

delta0 = 0 #-1e-2
delta_range = 0
num_turns=100
num_particles = 100_000

# delta0 = 1e-2
# delta_range = 0
# num_turns=20
# num_particles = 5_000_000

# # To see the different number of turns
# delta0 = 0e-2
# delta_range = 10e-3
# num_turns=5000
# num_particles = 5000

line = xt.Line.from_json(
    '../../test_data/psb_injection/line_and_particle.json')

# RF off!
tt = line.get_table()
ttcav = tt.rows[tt.element_type == 'Cavity']
for nn in ttcav.name:
    line.element_refs[nn].voltage=0

line.configure_bend_model(core='bend-kick-bend', edge='full')
line.twiss_default['method'] = '4d'

tw = line.twiss()
twom = line.twiss(delta0=delta0)
line.discard_tracker()

# Install dummy collective elements
s_sync = np.linspace(0, tw.circumference, 10)
for ii, ss in enumerate(s_sync):
    nn = f'sync_here_{ii}'
    line.insert(nn, obj=xt.Marker(), at=ss)
    line[nn].iscollective = True

import xtrack.synctime as st
st.install_sync_time_at_collective_elements(line)

import xobjects as xo
line.build_tracker(_context=xo.ContextCpu(omp_num_threads='auto'))

beta1 = tw.beta0 / 0.9

circumference = tw.circumference
zeta_min0 = -circumference/2*tw.beta0/beta1
zeta_max0 = circumference/2*tw.beta0/beta1


p = line.build_particles(
    delta=delta0 + delta_range * np.random.uniform(-1, 1, num_particles),
    x_norm=0, y_norm=0
)

# Need to take beta of actual particles to convert the distribution along the
# circumference to a distribution in time
p.zeta = (np.random.uniform(0, circumference, num_particles) / p.rvv * 0.999
          + (zeta_max0 - circumference) / p.rvv)

st.prepare_particles_for_sync_time(p, line)

p.y[(p.zeta > 1) & (p.zeta < 2)] = 1e-3  # kick
p.weight[(p.zeta > 5) & (p.zeta < 10)] *= 1.3

initial_histogram, z_init_hist = np.histogram(p.zeta, bins=200,
                                  range=(zeta_max0 - circumference, zeta_max0),
                                  weights=p.weight)

p0 = p.copy()

def particles(_, p):
    return p.copy()

def intensity(line, particles):
    return (np.sum(particles.weight[particles.state > 0])
                  / ((zeta_max0 - zeta_min0)/tw.beta0/clight))

def z_range(line, particles):
    mask_alive = particles.state > 0
    return particles.zeta[mask_alive].min(), particles.zeta[mask_alive].max()

def long_density(line, particles):
    mask_alive = particles.state > 0
    if not(np.any(particles.at_turn[mask_alive] == 0)): # don't check at the first turn
        assert np.all(particles.zeta[mask_alive] > zeta_min0)
        assert np.all(particles.zeta[mask_alive] < zeta_max0)
    return np.histogram(particles.zeta[mask_alive], bins=200,
                        range=(zeta_min0, zeta_max0),
                        weights=particles.weight[mask_alive])

def y_mean_hist(line, particles):

    mask_alive = particles.state > 0
    if not(np.any(particles.at_turn[mask_alive] == 0)): # don't check at the first turn
        assert np.all(particles.zeta[mask_alive] > zeta_min0)
        assert np.all(particles.zeta[mask_alive] < zeta_max0)
    return np.histogram(particles.zeta[mask_alive], bins=200,
                        range=(zeta_min0, zeta_max0), weights=particles.y[mask_alive])


line.enable_time_dependent_vars = True

line.track(p, num_turns=num_turns, log=xt.Log(intensity=intensity,
                                         long_density=long_density,
                                         y_mean_hist=y_mean_hist,
                                         z_range=z_range,
                                         particles=particles
                                         ), with_progress=2)

inten = line.log_last_track['intensity']

f_rev_ave = 1 / tw.T_rev0 * (1 - tw.slip_factor * p.delta.mean())
t_rev_ave = 1 / f_rev_ave

inten_exp =  np.sum(p0.weight) / t_rev_ave

z_axis = line.log_last_track['long_density'][0][1]
hist_mat = np.array([rr[0] for rr in line.log_last_track['long_density']])
hist_y = np.array([rr[0] for rr in line.log_last_track['y_mean_hist']])

dz = z_axis[1] - z_axis[0]
y_vs_t = np.fliplr(hist_y).flatten() # need to flip because of the minus in z = -beta0 c t
intensity_vs_t = np.fliplr(hist_mat).flatten()
z_unwrapped = np.arange(0, len(y_vs_t)) * dz
t_unwrapped = z_unwrapped / (tw.beta0 * clight)

z_range_size = z_axis[-1] - z_axis[0]
t_range_size = z_range_size / (tw.beta0 * clight)

import nafflib
f_harmons = nafflib.get_tunes(intensity_vs_t, N=50)[0] / (t_unwrapped[1] - t_unwrapped[0])
f_nominal = 1 / tw.T_rev0
dt_expected = -(twom.zeta[-1] - twom.zeta[0]) / tw.beta0 / clight
f_expected = 1 / (tw.T_rev0 + dt_expected)

f_measured = f_harmons[np.argmin(np.abs(f_harmons - f_nominal))]

print('f_nominal:  ', f_nominal, ' Hz')
print('f_expected: ', f_expected, ' Hz')
print('f_measured: ', f_measured, ' Hz')
print('Error:      ', f_measured - f_expected, 'Hz')

assert np.isclose(f_expected, f_measured, rtol=0, atol=2) # 2 Hz tolerance
assert np.isclose(np.mean(inten), inten_exp, rtol=1e-2, atol=0)
assert np.allclose(p.at_turn, num_turns*0.9, rtol=3e-2, atol=0) #beta1 defaults to 0.1

tt = line.get_table()
tt_synch = tt.rows[tt.element_type=='SyncTime']
assert len(tt_synch) == 12
assert tt_synch.name[0] == 'synctime_start'
assert tt_synch.name[-1] == 'synctime_end'
assert np.all(tt_synch.name[5] == 'synctime_4')
assert line['synctime_start'].at_start
assert not line['synctime_end'].at_start
assert not line['synctime_4'].at_start
assert line['synctime_end'].at_end
assert not line['synctime_start'].at_end
assert not line['synctime_4'].at_end

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(inten, label='xtrack')
plt.axhline(inten_exp, color='C1', label='expected')
plt.axhline(np.sum(p0.weight) / tw.T_rev0, color='C3', label='N/T_rev0')
plt.legend(loc='best')
plt.xlabel('Turn')
plt.ylim(inten_exp*0.95, inten_exp*1.05)

plt.figure(2)
plt.plot(p.delta, p.at_turn, '.')
plt.ylabel('Number of revolutions')
plt.xlabel(r'$\Delta P / P_0$')

plt.figure(3)
plt.plot([zz[1]-zz[0] for zz in line.log_last_track['z_range']])
plt.ylabel('z range [m]')
plt.xlabel('Turn')

plt.figure(4)
plt.plot(np.array([0.5*(zz[1] + zz[0]) for zz in line.log_last_track['z_range']]),
         label='z range center')
plt.plot([zz[0] for zz in line.log_last_track['z_range']], color='C1', linestyle='--', label='z range min')
plt.plot([zz[1] for zz in line.log_last_track['z_range']], color='C1', linestyle='--', label='z range max')
plt.plot()
plt.ylabel('z range[m]')
plt.legend(loc='best')
plt.xlabel('Turn')


plt.figure(5)
plt.pcolormesh(z_axis, np.arange(0, hist_mat.shape[0],1),
           hist_mat[:-1,:])


plt.figure(6)
plt.pcolormesh(z_axis, np.arange(0, hist_y.shape[0],1),
           hist_y[:-1,:])

plt.figure(7)
mask_alive = p.state>0
plt.plot(p.zeta[mask_alive], p.y[mask_alive], '.')
plt.axvline(x=circumference/2*tw.beta0/beta1, color='C1')
plt.axvline(x=-circumference/2*tw.beta0/beta1, color='C1')
plt.xlabel('z [m]')
plt.ylabel('x [m]')

f8 = plt.figure(8)
ax1 = plt.subplot(2, 1, 1)
plt.plot(t_unwrapped*1e6, y_vs_t, '-')
plt.ylabel('y mean [m]')
ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(t_unwrapped*1e6, intensity_vs_t/np.mean(intensity_vs_t), '-')
plt.ylabel('Beam line density [a.u.]')
plt.xlabel('t [us]')
plt.ylim(bottom=0)
for tt in np.arange(0, t_unwrapped[-1], 1/f_nominal):
    for ax in [ax1, ax2]:
        ax.axvline(x=tt*1e6, color='red', linestyle='--', alpha=0.5)

# zoom
ax1.set_xlim(0, 15)


plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import os
import sys
import json
import yaml
import numpy as np


#####################################################
# Read general configurations and setup envirnoment #
#####################################################

assert not(os.path.isfile('config.yaml') and os.path.isfile('config.py')), \
        "Please specify only a config file (yaml or py)"

try:
    with open('config.yaml','r') as fid:
        configuration = yaml.safe_load(fid)
except:
    from config import configuration

mode = configuration['mode']
tol_beta = configuration['tol_beta']
tol_sep = configuration['tol_sep']
flat_tol = configuration['tol_co_flatness']
links = configuration['links']
optics_file = configuration['optics_file']
check_betas_at_ips = configuration['check_betas_at_ips']
check_separations_at_ips = configuration['check_separations_at_ips']
save_intermediate_twiss = configuration['save_intermediate_twiss']
enable_lumi_control = configuration['enable_lumi_control']
enable_imperfections = configuration['enable_imperfections']
enable_crabs = configuration['enable_crabs']
match_q_dq_with_bb = configuration['match_q_dq_with_bb']
knob_settings = configuration['knob_settings']
knob_names = configuration['knob_names']


# Make links
for kk in links.keys():
    if os.path.exists(kk):
        os.remove(kk)
    os.symlink(os.path.abspath(links[kk]), kk)

# Create empty temp folder
os.system('rm -r temp')
os.system('mkdir temp')

# Execute customization script if present
os.system('bash customization.bash')

# Import pymask
sys.path.append('./modules')
import pymask as pm

# Import user-defined optics-specific tools
import optics_specific_tools as ost

######################################
# Check parameters and activate mode #
######################################

# Define configuration
(beam_to_configure, sequences_to_check, sequence_to_track, generate_b4_from_b2,
    track_from_b4_mad_instance, enable_bb_python, enable_bb_legacy,
    force_disable_check_separations_at_ips,
    ) = pm.get_pymask_configuration(mode)

if force_disable_check_separations_at_ips:
    check_separations_at_ips = False

if not(enable_crabs):
    knob_settings['par_crab1'] = 0.
    knob_settings['par_crab5'] = 0.

########################
# Build MAD-X instance #
########################

# Start mad
Madx = pm.Madxp
mad = Madx(command_log="mad_collider.log")

# Set verbose flag
mad.globals.par_verbose = int(configuration['verbose_mad_parts'])

# Build sequence (alse creates link to optics_toolkit and calls it)
ost.build_sequence(mad, beam=beam_to_configure)

# Set twiss formats for MAD-X parts (macro from opt. toolkit)
mad.input('exec, twiss_opt;')

# Apply optics
ost.apply_optics(mad, optics_file=optics_file)

# Attach beam to sequences
mad.globals.nrj = configuration['beam_energy_tot']
particle_type = 'proton'

if 'particle_mass' in configuration.keys():
    particle_mass = configuration['particle_mass']
    particle_type = 'ion'
else:
    particle_mass = mad.globals.pmass # proton mass

if 'particle_charge' in configuration.keys():
    particle_charge = configuration['particle_charge']
    particle_type = 'ion'
else:
    particle_charge = 1.

gamma_rel = (particle_charge*configuration['beam_energy_tot'])/particle_mass
for ss in mad.sequence.keys():
    # bv and bv_aux flags
    if ss == 'lhcb1':
        ss_beam_bv, ss_bv_aux = 1, 1
    elif ss == 'lhcb2':
        if int(beam_to_configure) == 4:
            ss_beam_bv, ss_bv_aux = 1, -1
        else:
            ss_beam_bv, ss_bv_aux = -1, 1

    mad.globals['bv_aux'] = ss_bv_aux
    mad.input(f'''
    beam, particle={particle_type},sequence={ss},
        energy={configuration['beam_energy_tot']*particle_charge},
        sigt={configuration['beam_sigt']},
        bv={ss_beam_bv},
        npart={configuration['beam_npart']},
        sige={configuration['beam_sige']},
        ex={configuration['beam_norm_emit_x'] * 1e-6 / gamma_rel},
        ey={configuration['beam_norm_emit_y'] * 1e-6 / gamma_rel},
        mass={particle_mass},
        charge={particle_charge};
    ''')


# Test machine before any change
twiss_dfs, other_data = ost.twiss_and_check(mad, sequences_to_check,
        tol_beta=tol_beta, tol_sep=tol_sep,
        twiss_fname='twiss_from_optics',
        save_twiss_files=save_intermediate_twiss,
        check_betas_at_ips=check_betas_at_ips,
        check_separations_at_ips=check_separations_at_ips)

# Set IP1-IP5 phase and store corresponding reference
mad.input("call, file='modules/submodule_01c_phase.madx';")

# Set optics-specific knobs
ost.set_optics_specific_knobs(mad, knob_settings, mode)

# Crossing-save and some reference measurements
mad.input('exec, crossing_save;')
mad.input("call, file='modules/submodule_01e_final.madx';")


#################################
# Check bahavior of orbit knobs #
#################################

# Check flat machine
mad.input('exec, crossing_disable;')
twiss_dfs, other_data = ost.twiss_and_check(mad, sequences_to_check,
        tol_beta=tol_beta, tol_sep=tol_sep,
        twiss_fname='twiss_no_crossing',
        save_twiss_files=save_intermediate_twiss,
        check_betas_at_ips=check_betas_at_ips, check_separations_at_ips=check_separations_at_ips)

# Check orbit flatness
for ss in twiss_dfs.keys():
    tt = twiss_dfs[ss]
    assert np.max(np.abs(tt.x)) < flat_tol
    assert np.max(np.abs(tt.y)) < flat_tol

# Check machine after crossing restore
mad.input('exec, crossing_restore;')
twiss_dfs, other_data = ost.twiss_and_check(mad, sequences_to_check,
        tol_beta=tol_beta, tol_sep=tol_sep,
        twiss_fname='twiss_with_crossing',
        save_twiss_files=save_intermediate_twiss,
        check_betas_at_ips=check_betas_at_ips, check_separations_at_ips=check_separations_at_ips)


#################################
# Set luminosity in IP2 and IP8 #
#################################

if len(sequences_to_check) == 2:
    print('Luminosities before leveling (crab cavities are not considered):')
    pm.print_luminosity(mad, twiss_dfs,
            configuration['nco_IP1'], configuration['nco_IP2'],
            configuration['nco_IP5'], configuration['nco_IP8'])
else:
    print('Warning: Luminosity computation requires two beams')


if not enable_lumi_control:
    print('Separations in IP2 and IP8 are left untouched')
elif enable_bb_legacy or mode=='b4_without_bb':
    mad.use(f'lhcb{beam_to_configure}')
    if mode=='b4_without_bb':
        print('Leveling not working in this mode!')
    else:
        if particle_type == 'ion': # the legacy macro for BB have been checked but not maintained
            raise ValueError
        # Luminosity levelling
        vars_for_legacy_level = ['lumi_ip8',
            'nco_IP1', 'nco_IP2', 'nco_IP5', 'nco_IP8']
        mad.set_variables_from_dict({
            'par_'+kk: configuration[kk] for kk in vars_for_legacy_level})
        mad.input("call, file='modules/module_02_lumilevel.madx';")
else:
    print('Start pythonic leveling:')
    ost.lumi_control(mad, twiss_dfs, configuration, knob_names)

# Force leveling
if 'force_leveling' in configuration.keys():
    force_leveling  = configuration['force_leveling']
    if force_leveling is not None:
        for kk in force_leveling.keys():
            mad.globals[kk] = force_leveling[kk]

# Re-save knobs (for the last time!)
mad.input('exec, crossing_save;')

# Check machine after leveling
mad.input('exec, crossing_restore;')
twiss_dfs, other_data = ost.twiss_and_check(mad, sequences_to_check,
        tol_beta=tol_beta, tol_sep=tol_sep,
        twiss_fname='twiss_after_leveling',
        save_twiss_files=save_intermediate_twiss,
        check_betas_at_ips=check_betas_at_ips,
        check_separations_at_ips=check_separations_at_ips)

if len(sequences_to_check) == 2:
    print('Luminosities after leveling (crab cavities are not considered):')
    pm.print_luminosity(mad, twiss_dfs,
            configuration['nco_IP1'], configuration['nco_IP2'],
            configuration['nco_IP5'], configuration['nco_IP8'])
else:
    print('Luminosity computation requires two beams')


#####################
# Force on_disp = 0 #
#####################

mad.globals.on_disp = 0.
# will be restored later


###################################
# Compute beam-beam configuration #
###################################

# Prepare bb dataframes
if enable_bb_python:
    bbconfig = configuration['beambeam_config']
    bb_dfs = pm.generate_bb_dataframes(mad,
        ip_names=['ip1', 'ip2', 'ip5', 'ip8'],
        harmonic_number=35640,
        numberOfLRPerIRSide=bbconfig['numberOfLRPerIRSide'],
        bunch_spacing_buckets=bbconfig['bunch_spacing_buckets'],
        numberOfHOSlices=bbconfig['numberOfHOSlices'],
        bunch_num_particles = bbconfig['bunch_num_particles'],
        bunch_particle_charge = bbconfig['bunch_particle_charge'],
        sigmaz_m=bbconfig['sigmaz_m'],
        z_crab_twiss=bbconfig['z_crab_twiss']*float(enable_crabs),
        remove_dummy_lenses=True)

    # Here the datafremes can be edited, e.g. to set bbb intensity

###################
# Generate beam 4 #
###################

if generate_b4_from_b2:
    mad_b4 = Madx(command_log="mad_b4.log")
    ost.build_sequence(mad_b4, beam=4)

    pm.configure_b4_from_b2(mad_b4, mad)

    twiss_dfs_b2, other_data_b2 = ost.twiss_and_check(mad,
            sequences_to_check=['lhcb2'],
            tol_beta=tol_beta, tol_sep=tol_sep,
            twiss_fname='twiss_b2_for_b4check',
            save_twiss_files=save_intermediate_twiss,
            check_betas_at_ips=check_betas_at_ips, check_separations_at_ips=False)

    twiss_dfs_b4, other_data_b4 = ost.twiss_and_check(mad_b4,
            sequences_to_check=['lhcb2'],
            tol_beta=tol_beta, tol_sep=tol_sep,
            twiss_fname='twiss_b4_for_b4check',
            save_twiss_files=save_intermediate_twiss,
            check_betas_at_ips=check_betas_at_ips, check_separations_at_ips=False)

# For B1, to be generalized for B4
if 'filling_scheme_json' in configuration['beambeam_config'].keys():
    assert 'b4' not in mode
    filling_scheme_json = configuration['beambeam_config']['filling_scheme_json']
    bunch_to_track = configuration['beambeam_config']['bunch_to_track']
    bb_schedule_to_track_b1 = ost.create_bb_shedule_to_track(
                              filling_scheme_json,bunch_to_track, beam=1)
    bb_dfs['b1']=ost.filter_bb_df(bb_dfs['b1'],bb_schedule_to_track_b1)

##################################################
# Select mad instance for tracking configuration #
##################################################

# We will be working exclusively on the sequence to track
# Select mad object
if track_from_b4_mad_instance:
    mad_track = mad_b4
else:
    mad_track = mad

mad_collider = mad
del(mad)

# Twiss machine to track
twiss_dfs, other_data = ost.twiss_and_check(mad_track, sequences_to_check,
        tol_beta=tol_beta, tol_sep=tol_sep,
        twiss_fname='twiss_track_intermediate',
        save_twiss_files=save_intermediate_twiss,
        check_betas_at_ips=check_betas_at_ips, check_separations_at_ips=False)


#####################
# Install bb lenses #
#####################

# Python approach
if enable_bb_python:
    if track_from_b4_mad_instance:
        bb_df_track = bb_dfs['b4']
        assert(sequence_to_track=='lhcb2')
    else:
        bb_df_track = bb_dfs['b1']
        assert(sequence_to_track=='lhcb1')

    pm.install_lenses_in_sequence(mad_track, bb_df_track, sequence_to_track)

    # Disable bb (to be activated later)
    mad_track.globals.on_bb_charge = 0
else:
    bb_df_track = None

# Legacy bb macros
if enable_bb_legacy:
    bbconfig = configuration['beambeam_config']
    assert(beam_to_configure == 1)
    assert(not(track_from_b4_mad_instance))
    assert(not(enable_bb_python))
    mad_track.globals['par_on_bb_switch'] = 1
    mad_track.set_variables_from_dict(
       params=configuration['pars_for_legacy_bb_macros'])
    mad_track.set_variables_from_dict(
            params={f'par_nho_ir{ir}': bbconfig['numberOfHOSlices']
            for ir in [1,2,5,8]})
    mad_track.input("call, file='modules/module_03_beambeam.madx';")


#########################
# Install crab cavities #
#########################

if enable_crabs:
    mad_track.input("call, file='optics_toolkit/enable_crabcavities.madx';")
    # They are left off, they will be swiched on at the end:
    mad_track.globals.on_crab1 = 0
    mad_track.globals.on_crab5 = 0


##############################################
# Save references for tuning and corrections #
##############################################

mad_track.input("call, file='modules/submodule_04_1b_save_references.madx';")


#####################
# Force on_disp = 0 #
#####################

mad_track.globals.on_disp = 0.
# will be restored later


#############
# Final use #
#############

mad_track.use(sequence_to_track)
# Disable use
mad_track._use = mad_track.use
mad_track.use = None


##############################
# Install and correct errors #
##############################

if enable_imperfections:
    mad_track.set_variables_from_dict(
            configuration['pars_for_imperfections'])
    mad_track.input("call, file='modules/module_04_errors.madx';")
else:
    # Synthesize knobs
    mad_track.input('call, file="modules/submodule_04a_s1_prepare_nom_twiss_table.madx";')
    if configuration['enable_knob_synthesis']:
        mad_track.input('exec, crossing_disable;')
        mad_track.input("call, file='modules/submodule_04e_s1_synthesize_knobs.madx';")
    mad_track.input('exec, crossing_restore;')


##################
# Machine tuning #
##################

# Enable bb for matchings
if match_q_dq_with_bb:
    mad_track.globals['on_bb_charge'] = 1
else:
    mad_track.globals['on_bb_charge'] = 0

# Switch on octupoles
brho = mad_track.globals.nrj*1e9/mad_track.globals.clight
i_oct = configuration['oct_current']
beam_str = {'lhcb1':'b1', 'lhcb2':'b2'}[sequence_to_track]
for ss in '12 23 34 45 56 67 78 81'.split():
   mad_track.input(f'kof.a{ss}{beam_str} = kmax_mo*({i_oct})/imax_mo/({brho});')
   mad_track.input(f'kod.a{ss}{beam_str} = kmax_mo*({i_oct})/imax_mo/({brho});')


# Correct linear coupling
qx_fractional, qx_integer = np.modf(configuration['qx0'])
qy_fractional, qy_integer = np.modf(configuration['qy0'])
coupl_corr_info = pm.coupling_correction(mad_track,
        n_iterations=configuration['N_iter_coupling'],
        qx_integer=qx_integer, qy_integer=qy_integer,
        qx_fractional=qx_fractional, qy_fractional=qy_fractional,
        tune_knob1_name=knob_names['qknob_1'][sequence_to_track],
        tune_knob2_name=knob_names['qknob_2'][sequence_to_track],
        cmr_knob_name=knob_names['cmrknob'][sequence_to_track],
        cmi_knob_name=knob_names['cmiknob'][sequence_to_track],
        sequence_name=sequence_to_track, skip_use=True)

# Add custom values to coupling knobs
mad_track.globals[knob_names['cmrknob'][sequence_to_track]] += configuration['delta_cmr']
mad_track.globals[knob_names['cmiknob'][sequence_to_track]] += configuration['delta_cmi']

# Check strength limits
if enable_imperfections:
    mad_track.input('call, file="errors/HL-LHC/corr_limit.madx";')

# Rematch the orbit at IPs
mad_track.input("call, file='tools/rematchCOIP.madx';")

# Rematch the CO in the arc for dispersion correction
if mad_track.globals.on_disp != 0:
    mad_track.input("call, file='tools/rematchCOarc.madx';")

# Match tunes and chromaticities
pm.match_tune_and_chromaticity(mad_track,
        q1=configuration['qx0'],
        q2=configuration['qy0'],
        dq1=configuration['chromaticity_x'],
        dq2=configuration['chromaticity_y'],
        tune_knob1_name=knob_names['qknob_1'][sequence_to_track],
        tune_knob2_name=knob_names['qknob_2'][sequence_to_track],
        chromaticity_knob1_name=knob_names['chromknob_1'][sequence_to_track],
        chromaticity_knob2_name=knob_names['chromknob_2'][sequence_to_track],
        sequence_name=sequence_to_track, skip_use=True)

# Check strength limits
if enable_imperfections:
    mad_track.input("call, file='errors/HL-LHC/corr_value_limit.madx';")

# Switch on bb lenses
mad_track.globals.on_bb_charge = 1.

# Switch on RF cavities
mad_track.globals['vrf400'] = configuration['vrf_total']
if sequence_to_track == 'lhcb1':
    mad_track.globals['lagrf400.b1'] = 0.5
elif sequence_to_track == 'lhcb2':
    mad_track.globals['lagrf400.b2'] = 0.

# Switch on crab cavities
if enable_crabs:
    mad_track.globals.on_crab1 = knob_settings['on_crab1']
    mad_track.globals.on_crab5 = knob_settings['on_crab5']


#####################
# Generate sixtrack #
#####################

if enable_bb_legacy:
    mad_track.input("call, file='modules/module_06_generate.madx'")
else:
    pm.generate_sixtrack_input(mad_track,
        seq_name=sequence_to_track,
        bb_df=bb_df_track,
        output_folder='./',
        reference_num_particles_sixtrack=(
            mad_track.sequence[sequence_to_track].beam.npart),
        reference_particle_charge_sixtrack=mad_track.sequence[sequence_to_track].beam.charge,
        emitnx_sixtrack_um=(
            mad_track.sequence[sequence_to_track].beam.exn),
        emitny_sixtrack_um=(
            mad_track.sequence[sequence_to_track].beam.eyn),
        sigz_sixtrack_m=(
            mad_track.sequence[sequence_to_track].beam.sigt),
        sige_sixtrack=(
            mad_track.sequence[sequence_to_track].beam.sige),
        ibeco_sixtrack=1,
        ibtyp_sixtrack=0,
        lhc_sixtrack=2,
        ibbc_sixtrack=0,
        radius_sixtrack_multip_conversion_mad=0.017,
        skip_mad_use=True)


#######################################
# Save optics and orbit at start ring #
#######################################

optics_and_co_at_start_ring_from_madx = pm.get_optics_and_orbit_at_start_ring(
        mad_track, sequence_to_track, skip_mad_use=True)
with open('./optics_orbit_at_start_ring_from_madx.json', 'w') as fid:
    json.dump(optics_and_co_at_start_ring_from_madx, fid, cls=pm.JEncoder)

########################
# Generate xtrack line #
########################
if enable_bb_legacy:
    print('xtrack line is not generated with bb legacy macros')
else:
    pm.generate_xsuite_line(mad_track, sequence_to_track, bb_df_track,
                    optics_and_co_at_start_ring_from_madx,
                    folder_name = './xsuite_lines',
                    skip_mad_use=True,
                    prepare_line_for_xtrack=True)

###################################
#         Save final twiss        #
###################################

mad_track.globals.on_bb_charge = 0
mad_track.twiss()
tdf = mad_track.get_twiss_df('twiss')
sdf = mad_track.get_summ_df('summ')
tdf.to_parquet('final_twiss_BBOFF.parquet')
sdf.to_parquet('final_summ_BBOFF.parquet')


mad_track.globals.on_bb_charge = 1
mad_track.twiss()
tdf = mad_track.get_twiss_df('twiss')
sdf = mad_track.get_summ_df('summ')
tdf.to_parquet('final_twiss_BBON.parquet')
sdf.to_parquet('final_summ_BBON.parquet')

#############################    
#  Save sequence and errors #
#############################
# N.B. this erases the errors in the mad_track instance
pm.save_mad_sequence_and_error(mad_track, sequence_to_track, filename='final')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt
import xfields as xf

############
# Settings #
############

fname_line = ('../../test_data/sps_w_spacecharge/'
                  'line_no_spacecharge_and_particle.json')

bunch_intensity = 1e11/3 # Need short bunch to avoid bucket non-linearity
                         # to compare frozen/quasi-frozen and PIC
sigma_z = 22.5e-2/3
nemitt_x=2.5e-6
nemitt_y=2.5e-6
n_part=int(1e6)
num_turns=32

num_spacecharge_interactions = 540

# Available modes: frozen/quasi-frozen/pic
# mode = 'pic'
mode = 'quasi-frozen'

# Choose solver between `FFTSolver2p5DAveraged` and `FFTSolver2p5D`
pic_solver = 'FFTSolver2p5DAveraged'

####################
# Choose a context #
####################

context = xo.ContextCpu()
#context = xo.ContextCupy()
#context = xo.ContextPyopencl('0.0')

print(context)

#############
# Load line #
#############

with open(fname_line, 'r') as fid:
     input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xt.Particles.from_dict(input_data['particle'])


line.build_tracker(_context=context, compile=False)

#############################################
# Install spacecharge interactions (frozen) #
#############################################

lprofile = xf.LongitudinalProfileQGaussian(
        number_of_particles=bunch_intensity,
        sigma_z=sigma_z,
        z0=0.,
        q_parameter=1.)

xf.install_spacecharge_frozen(line=line,
                   longitudinal_profile=lprofile,
                   nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                   sigma_z=sigma_z,
                   num_spacecharge_interactions=num_spacecharge_interactions,
                   )

#################################
# Switch to PIC or quasi-frozen #
#################################

if mode == 'frozen':
    pass # Already configured in line
elif mode == 'quasi-frozen':
    xf.replace_spacecharge_with_quasi_frozen(
                                    line,
                                    update_mean_x_on_track=True,
                                    update_mean_y_on_track=True)
elif mode == 'pic':
    pic_collection, all_pics = xf.replace_spacecharge_with_PIC(
        line=line,
        n_sigmas_range_pic_x=8,
        n_sigmas_range_pic_y=8,
        nx_grid=256, ny_grid=256, nz_grid=100,
        n_lims_x=7, n_lims_y=3,
        z_range=(-3*sigma_z, 3*sigma_z),
        solver=pic_solver)
else:
    raise ValueError(f'Invalid mode: {mode}')

#################
# Build Tracker #
#################

line.build_tracker(_context=context)

######################
# Generate particles #
######################

# Build a line without spacecharge (recycling the track kernel)
line_sc_off = line.filter_elements(exclude_types_starting_with='SpaceCh')
line_sc_off.build_tracker(track_kernel=line.tracker.track_kernel)

# (we choose to match the distribution without accounting for spacecharge)
particles = xp.generate_matched_gaussian_bunch(line=line_sc_off,
         num_particles=n_part, total_intensity_particles=bunch_intensity,
         nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z)

####################################################
# Check that everything is on the selected context #
####################################################

assert particles._context == context
assert len(set([ee._buffer for ee in line.elements])) == 1 # All on same context
assert line._context == context

#########
# Track #
#########
line.track(particles, num_turns=num_turns, with_progress=1)




import xtrack as xt
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=2e9, mass0=xt.PROTON_MASS_EV)

sub_line = env.new_line(components=[
    env.new('mid_subline', xt.Marker, at=1.0),
    env.new('mq1_subline', xt.Quadrupole, length=0.3, at=-0.5, from_='mid_subline'),
    env.new('mq2_subline', xt.Quadrupole, length=0.3, at=0.5, from_='mid_subline')])

line = env.new_line(components=[
    env.new('bb', xt.Bend, length=1.0, at=2),
    env.place(sub_line, at=5.0),
    env.place(sub_line, at=8.0),
    ])

line.get_table()
# is:
# Table: 17 rows, 8 cols
# name                       s element_type isthick isreplica parent_name ...
# drift_4                    0 Drift           True     False None
# bb                       1.5 Bend            True     False None
# drift_5                  2.5 Drift           True     False None
# drift_1::0             4.175 Drift           True     False None
# mq1_subline::0         4.525 Quadrupole      True     False None
# drift_2::0             4.825 Drift           True     False None
# mid_subline::0         5.175 Marker         False     False None
# drift_3::0             5.175 Drift           True     False None
# mq2_subline::0         5.525 Quadrupole      True     False None
# drift_6                5.825 Drift           True     False None
# drift_1::1             7.175 Drift           True     False None
# mq1_subline::1         7.525 Quadrupole      True     False None
# drift_2::1             7.825 Drift           True     False None
# mid_subline::1         8.175 Marker         False     False None
# drift_3::1             8.175 Drift           True     False None
# mq2_subline::1         8.525 Quadrupole      True     False None
# _end_point             8.825                False     False None



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from .mathlibs import MathlibDefault

from .temp_pyparticles import Pyparticles

class TestParticles(Pyparticles):
    _m = MathlibDefault



import xtrack as xt
env = xt.get_environment(verbose=True)

env.vars.default_to_zero = True

env.new('qf1', 'Quadrupole', length=2*0.175, k1='qf1k1')
env.new('qd',  'Quadrupole', length=2*0.175, k1='qd1k1')
env.new('qf2', 'Quadrupole', length=2*0.175, k1='qf2k1')
env.new('qa',  'Quadrupole', length=0.556,   k1='qa1k1')

env.new('mb',  'Bend', length=1.661, angle=0.3926990817,
        edge_entry_angle=0.19634954085, edge_exit_angle=0.19634954085)

env.new('xcd1', 'Sextupole', length=0.2, k2='k2xcd')
env.new('xcf1', 'Sextupole', length=0.2, k2='k2xcf')
env.new('xcd2', 'Sextupole', length=0.2, k2='k2xcd')
env.new('xcf2', 'Sextupole', length=0.2, k2='k2xcf')
env.new('xrra', 'Sextupole', length=0.2, k2='k2xrr_a')
env.new('xrrb', 'Sextupole', length=0.2, k2='k2xrr_b')

env.new('es_injection', 'Marker')
env.new('es_marker',    'Marker')
env.new('betkick',      'Marker')
env.new('rfkokick',     'Marker')
env.new('pimm_start',   'Marker')

env.new('pimm_cavity',  'Cavity')

pimm = env.new_builder(name='pimm')

# Place elements
pimm.new('end_ring', 'Marker', at=75.24) # defines ring length
pimm.place('pimm_start', at=0)
pimm.place('pimm_cavity', at=0.001)
pimm.place('es_injection', at=1.05)
pimm.place('qf1', at=2.3875)
pimm.place('mb', at=3.8125)
pimm.place('qd', at=5.2925)
pimm.place('mb', at=7.0475)
pimm.place('qf1', at=8.3275)
pimm.place('rfkokick', at= 9.30875)
pimm.place('xcd1', at=10.29)
pimm.place('qf2', at=10.6775)
pimm.place('mb', at=12.1325)
pimm.place('qd', at=14.2625)
pimm.place('mb', at=15.9175)
pimm.place('qf2', at=17.1975)
pimm.place('betkick', at=18.7225)
pimm.place('qf2', at=20.2475)
pimm.place('mb', at=21.7025)
pimm.place('qd', at=23.1825)
pimm.place('xcf1', at=24.045)
pimm.place('mb', at=25.4875)
pimm.place('qf2', at=26.7675)
pimm.place('qa', at=27.9425)
pimm.place('qf1', at=29.1175)
pimm.place('mb', at=30.5725)
pimm.place('qd', at=32.1525)
pimm.place('mb', at=33.8075)
pimm.place('qf1', at=35.0575)
pimm.place('xrra', at=35.0575 + 0.6)
pimm.place('xrrb', at=40.0075 - 0.6)
pimm.place('qf1', at=40.0075)
pimm.place('mb', at=41.4325)
pimm.place('qd', at=42.9125)
pimm.place('mb', at=44.6675)
pimm.place('qf1', at=45.9475)
pimm.place('xcd2', at=47.91)
pimm.place('qf2', at=48.2975)
pimm.place('mb', at=49.7525)
pimm.place('qd', at=51.8825)
pimm.place('mb', at=53.5375)
pimm.place('qf2', at=54.8175)
pimm.place('qf2', at=57.8675)
pimm.place('mb', at=59.3225)
pimm.place('qd', at=60.8025)
pimm.place('xcf2', at=61.665)
pimm.place('mb', at=63.1075)
pimm.place('qf2', at=64.3875)
pimm.place('qf1', at=66.7375)
pimm.place('mb', at=68.1925)
pimm.place('qd', at=69.7725)
pimm.place('mb', at=71.4275)
pimm.place('qf1', at=72.6775)
pimm.place('es_marker', at=73.25225)

pimm = pimm.build() # Becomes a line
env.default_to_zero = False


import xtrack as xt

# Load a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

class ActionMeasAmplDet(xt.Action):

    def __init__(self, line, num_turns, nemitt_x, nemitt_y):

        self.line = line
        self.num_turns = num_turns
        self.nemitt_x = nemitt_x
        self.nemitt_y = nemitt_y

    def run(self):

        det_coefficients = self.line.get_amplitude_detuning_coefficients(
                                nemitt_x=self.nemitt_x, nemitt_y=self.nemitt_y,
                                num_turns=self.num_turns)

        out = {'d_xx': det_coefficients['det_xx'],
               'd_yy': det_coefficients['det_yy']}

        return out

action = ActionMeasAmplDet(line=line, nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                           num_turns=128)

opt = line.match(vary=xt.VaryList(['kof.a23b1', 'kod.a23b1'], step=1.),
                 targets=[action.target('d_xx', 1000., tol=0.1),
                          action.target('d_yy', 2000., tol=0.1)])

opt.target_status()
# prints:
#
# Target status:
# id state tag tol_met     residue current_val target_val description
#  0 ON           True   0.0844456     1000.08       1000 'd_xx', val=1000, ...
#  1 ON           True -0.00209987        2000       2000 'd_yy', val=2000, ...